"""
fprime.fbuild.settings:

An implementation used to pull settings into the fprime build. This version uses INI files in order
to load the settings from the settings.default file that is part of the F prime project directory.

@author mstarch
"""

import configparser
import os
import sys
from enum import Enum
from functools import partial
from pathlib import Path
from typing import Any, Callable, Dict, List, Union


class EnvironmentVariableInterpolation(configparser.BasicInterpolation):
    """Interpolation for environment variables embedded in setting.ini

    settings.ini will extrapolate $/${} as an environment variable. This will allow basic environment variable
    replacement. It is illegal to supply an environment variable with a $ or %( as part of the value because that might
    trigger unanticipated recursive parsing.

    Note: environment variable substitution is only performed in the environment section and nowhere else.

    Based off: https://stackoverflow.com/questions/26586801/configparser-and-string-interpolation-with-env-variable
    """

    def before_get(self, parser, section, option, value, defaults):
        """Pre-process sections to replace environment variables

        Runs before the value is gotten to replace environment variables. It will use os.path.expandvars to do the
        actual substitution on any value in the "environment" section. Other pre-processing is done first.

        Args:
            parser: unused
            section: environment substitution will be done only when set to "environment"
            option: unused
            value: will be searched for and replace environment variables
            defaults: unused
        Returns:
            the value post substitution
        """
        value = super().before_get(parser, section, option, value, defaults)
        return os.path.expandvars(value) if section == "environment" else value


class SettingType(Enum):
    """Designates the type of the setting"""

    PATH = 0
    PATH_LIST = 1
    STRING = 2


def find_fprime(settings: dict) -> Path:
    """
    Finds F prime by recursing parent to parent until a matching directory is found.
    """
    needle = Path("fprime/cmake/FPrime.cmake")
    path = settings["_cmake_project_root"]
    while path != path.parent:
        if (path / needle).is_file():
            return path / "fprime"
        path = path.parent
    raise FprimeLocationUnknownException(
        "Please set 'framework_path' in [fprime] section in 'settings.ini"
    )


def join(key: Path, addition: str, settings: dict):
    """Joins a settings key to the addition"""
    return settings[key] / addition


class IniSettings:
    """Class to load settings from INI files"""

    DEF_FILE = "settings.ini"
    SET_ENV = "FPRIME_SETTINGS_FILE"

    FPRIME_FIELDS = [
        ("framework_path", SettingType.PATH, find_fprime),
        ("project_root", SettingType.PATH, lambda settings: settings["framework_path"]),
        ("default_toolchain", SettingType.STRING, "native"),
        ("default_ut_toolchain", SettingType.STRING, "native"),
        ("library_locations", SettingType.PATH_LIST, []),
        ("component_cookiecutter", SettingType.STRING, "default"),
        ("deployment_cookiecutter", SettingType.STRING, "default"),
    ]

    PLATFORM_FIELDS = [
        (
            "config_directory",
            SettingType.PATH,
            partial(join, "framework_path", "config"),
        ),
        (
            "install_destination",
            SettingType.PATH,
            partial(join, "_cmake_project_root", "build-artifacts"),
        ),
        (
            "environment_file",
            SettingType.PATH,
            lambda settings: settings["settings_file"],
        ),
        ("default_cmake_options", SettingType.STRING, ""),
    ]

    @staticmethod
    def read_safe_path(
        parser: configparser.ConfigParser,
        section: str,
        key: str,
        ini_file: Path,
        exists: bool = True,
    ) -> List[Path]:
        """
        Reads path(s), safely, from the config parser.  Validates the path(s) exists or raises an exception. Paths are
        separated by ':'.  This will also expand relative paths relative to the settings file.

        :param parser: parser to read from
        :param section: section to read from
        :param key: key to read from
        :param ini_file: ini_file path for clean errors
        :param exists: expect the path to exist. Default: True, must exist.
        :return: path, validated
        """
        base_dir = os.path.dirname(ini_file)
        all_paths = parser.get(section, key, fallback="").split(":")
        expanded = []
        for path in all_paths:
            if not path:
                continue
            full_path = os.path.abspath(os.path.normpath(os.path.join(base_dir, path)))
            if exists and not os.path.exists(full_path):
                msg = f"Nonexistent path '{path}' found in section '{section}' option '{key}' of file '{ini_file}'"
                raise FprimeSettingsException(msg)
            expanded.append(Path(full_path).resolve())
        return expanded

    @staticmethod
    def read_setting(
        config_parser: configparser.ConfigParser,
        settings: Dict[str, Any],
        section: str,
        key: str,
        settings_type: SettingType,
        default: Union[Callable, Any],
    ):
        """Reads an individual setting"""

        def get_default_value():
            """Calculates the default value for the given setting"""
            return default(settings) if callable(default) else default

        if config_parser is None:
            value = get_default_value()
        elif settings_type == SettingType.STRING:
            value = config_parser.get(section, key, fallback=get_default_value())
        elif settings_type == SettingType.PATH:
            paths_list = IniSettings.read_safe_path(
                config_parser,
                section,
                key,
                settings["settings_file"],
                key != "install_destination",
            )
            value = paths_list[0] if paths_list else get_default_value()
        elif settings_type == SettingType.PATH_LIST:
            paths_list = IniSettings.read_safe_path(
                config_parser,
                section,
                key,
                settings["settings_file"],
                key != "install_destination",
            )
            value = paths_list or get_default_value()
        else:
            raise FprimeSettingsException("Invalid settings specification")
        return value

    @staticmethod
    def load(settings_file: Path, platform: str = "native", is_ut: bool = False):
        """
        Load settings from specified file or from specified build directory. Either a specific file or the build
        directory must be not None.

        :param settings_file: file to load settings from (in INI format). Must be specified if build_dir is not.
        :param platform: platform to read platform specific settings
        :param is_ut: is this a unit test build
        :return: a dictionary of needed settings
        """
        settings_file = (
            Path.cwd() / IniSettings.DEF_FILE
            if settings_file is None
            else settings_file
        ).resolve()
        # Setup a config parser, or none if the settings file does not exist
        confparse = None
        if settings_file.exists():
            confparse = configparser.ConfigParser(
                interpolation=EnvironmentVariableInterpolation()
            )
            confparse.read(settings_file)
        else:
            print(f"[WARNING] {settings_file} does not exist", file=sys.stderr)

        settings = {
            "settings_file": settings_file,
            "_cmake_project_root": settings_file.parent,
        }

        # Read fprime and platform settings from the "fprime" section
        for key, settings_type, default in (
            IniSettings.FPRIME_FIELDS + IniSettings.PLATFORM_FIELDS
        ):
            settings[key] = IniSettings.read_setting(
                confparse, settings, "fprime", key, settings_type, default
            )

        # Calculate the platform if not specified
        if not platform or platform == "default":
            platform = (
                settings["default_ut_toolchain"]
                if is_ut
                else settings["default_toolchain"]
            )

        # Read platform settings overtop of fprime settings
        for key, settings_type, default in IniSettings.PLATFORM_FIELDS:
            settings[key] = IniSettings.read_setting(
                confparse,
                settings,
                platform,
                key,
                settings_type,
                settings.get(key, default),
            )

        settings["environment"] = IniSettings.load_environment(
            settings["environment_file"]
        )
        del settings["_cmake_project_root"]

        # add _fprime_packages to library locations
        try:
            if os.path.exists(settings["project_root"] / "_fprime_packages"):
                # glob all folders
                for folder in os.listdir(settings["project_root"] / "_fprime_packages"):
                    settings["library_locations"].append(
                        Path(settings["project_root"] / "_fprime_packages" / folder)
                    )
        except FileNotFoundError:
            # we shouldn't error out if the _fprime_packages folder doesn't exist
            pass

        return settings

    @staticmethod
    def load_environment(env_file):
        """
        Load the environment from the given parser.

        :param env_file: load environment from this file
        :return: environment dictionary
        """
        parser = configparser.ConfigParser(
            interpolation=EnvironmentVariableInterpolation()
        )
        parser.optionxform = str
        parser.read(env_file)
        env_dict = {}
        try:
            for key, value in parser.items("environment"):
                env_dict[key] = value
        except configparser.NoSectionError:
            pass  # Ignore missing environment
        return env_dict


class FprimeLocationUnknownException(Exception):
    """Fprime location could not be determined"""


class FprimeSettingsException(Exception):
    """An exception for handling F prime settings misconfiguration"""



"""
Tests for fprime.fpp.common
"""

from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from fprime.fpp.common import FppUtility, FppMissingSupportFiles


@pytest.fixture
def mock_build(tmp_path):
    """Pytest fixture for a mock build object"""
    build = MagicMock()
    build.build_dir = tmp_path
    build.get_build_cache_path.return_value = tmp_path / "cache"
    (tmp_path / "cache").mkdir()
    return build


def test_init():
    """Test FppUtility initialization"""
    utility = FppUtility("fpp-to-dict")
    assert utility.utility == "fpp-to-dict"
    assert utility.imports_as_sources is True

    utility_no_imports = FppUtility("fpp-check", imports_as_sources=False)
    assert utility_no_imports.imports_as_sources is False


def test_get_locations_file(mock_build, tmp_path):
    """Test get_locations_file successfully finds file"""
    expected_path = tmp_path / "locs.fpp"
    expected_path.touch()
    path = FppUtility.get_locations_file(mock_build)
    assert path == expected_path


def test_get_locations_file_missing(mock_build):
    """Test get_locations_file fails when file is missing"""
    with pytest.raises(FppMissingSupportFiles):
        FppUtility.get_locations_file(mock_build)


def test_get_fpp_inputs(mock_build, tmp_path):
    """Test get_fpp_inputs successfully reads files"""
    import_list_file = tmp_path / "cache" / "fpp-import-list"
    source_list_file = tmp_path / "cache" / "fpp-source-list"

    import_list_file.write_text("A.fpp;B.fpp;")
    source_list_file.write_text("C.fpp;D.fpp;")

    imports, sources = FppUtility.get_fpp_inputs(mock_build, tmp_path)

    assert imports == [Path("A.fpp"), Path("B.fpp")]
    assert sources == [Path("C.fpp"), Path("D.fpp")]


def test_get_fpp_inputs_missing(mock_build, tmp_path):
    """Test get_fpp_inputs fails when a file is missing"""
    with pytest.raises(FppMissingSupportFiles):
        FppUtility.get_fpp_inputs(mock_build, tmp_path)


@patch("subprocess.run")
@patch("shutil.which", return_value="/path/to/fpp-to-dict")
def test_execute_imports_as_sources(mock_which, mock_run, mock_build, tmp_path):
    """Test execute with imports as sources"""
    # Setup
    (tmp_path / "locs.fpp").touch()
    (tmp_path / "cache" / "fpp-import-list").write_text("A.fpp;B.fpp;")
    (tmp_path / "cache" / "fpp-source-list").write_text("C.fpp;D.fpp;")

    utility = FppUtility("fpp-to-dict", imports_as_sources=True)
    utility.execute(mock_build, tmp_path, ({}, []))

    # Check call
    mock_run.assert_called_once()
    call_args = mock_run.call_args[0][0]
    assert call_args == [
        "fpp-to-dict",
        str(tmp_path / "locs.fpp"),
        "A.fpp",
        "B.fpp",
        "C.fpp",
        "D.fpp",
    ]


@patch("subprocess.run")
@patch("shutil.which", return_value="/path/to/fpp-check")
def test_execute_imports_as_flag(mock_which, mock_run, mock_build, tmp_path):
    """Test execute with imports as -i flag"""
    # Setup
    (tmp_path / "locs.fpp").touch()
    (tmp_path / "cache" / "fpp-import-list").write_text("A.fpp;B.fpp;")
    (tmp_path / "cache" / "fpp-source-list").write_text("C.fpp;D.fpp;")

    utility = FppUtility("fpp-check", imports_as_sources=False)
    utility.execute(mock_build, tmp_path, ({}, []))

    # Check call
    mock_run.assert_called_once()
    call_args = mock_run.call_args[0][0]
    assert call_args == [
        "fpp-check",
        "-i",
        "A.fpp,B.fpp",
        str(tmp_path / "locs.fpp"),
        "C.fpp",
        "D.fpp",
    ]



import fprime.fpp.utils.fpp_to_json.node_structs as NodeStructs
import fprime.fpp.utils.fpp_to_json.helpers as Utils


class ModuleConverter:
    """
    AST Name: DefModule
    """

    def __init__(self, module: NodeStructs.Module):
        self.module_struct: NodeStructs.Module = module

        self.module_struct.preannot = module.ast[0]
        self.module_JSON = module.ast[1]
        self.module_struct.postannot = module.ast[2]

    def convert(self):
        self.module_struct.name = self.module_JSON["DefModule"]["node"]["AstNode"][
            "data"
        ]["name"]
        self.module_struct.qf = self.module_struct.name

        self.module_struct.members = self.module_JSON["DefModule"]["node"]["AstNode"][
            "data"
        ]["members"]

        return self.module_struct


class ConstantConverter:
    """
    AST Name: DefConstant
    """

    def __init__(self, constant: NodeStructs.Constant):
        self.constant_struct: NodeStructs.Constant = constant

        self.constant_struct.preannot = constant.ast[0]
        self.constant_JSON = constant.ast[1]
        self.constant_struct.postannot = constant.ast[2]

    def convert(self):
        self.constant_struct.id = self.constant_JSON["DefConstant"]["node"]["AstNode"][
            "data"
        ]["name"]
        self.constant_struct.qf = self.constant_struct.id

        constant_Value_JSON = self.constant_JSON["DefConstant"]["node"]["AstNode"][
            "data"
        ]["value"]

        self.constant_struct.value = Utils.parse_constant(constant_Value_JSON)

        return self.constant_struct


class InstanceSpecConverter:
    """
    AST Name: SpecCompInstance
    """

    def __init__(self, instance: NodeStructs.InstanceSpec):
        self.instance_struct: NodeStructs.InstanceSpec = instance

        self.instance_struct.preannot = instance.ast[0]
        self.instance_JSON = instance.ast[1]
        self.instance_struct.postannot = instance.ast[2]

    def convert(self):
        instance_bit_1 = self.instance_JSON["SpecCompInstance"]["node"]["AstNode"][
            "data"
        ]["instance"]

        self.instance_struct.name = Utils.value_parser(instance_bit_1)
        self.instance_struct.qf = self.instance_struct.name

        if (
            "Public"
            in self.instance_JSON["SpecCompInstance"]["node"]["AstNode"]["data"][
                "visibility"
            ]
        ):
            self.instance_struct.visibility = ""
        elif (
            "Private"
            in self.instance_JSON["SpecCompInstance"]["node"]["AstNode"]["data"][
                "visibility"
            ]
        ):
            self.instance_struct.visibility = "private"

        return self.instance_struct


class CompInstanceConverter:
    """
    AST Name: DefComponentInstance
    """

    def __init__(self, instance: NodeStructs.ComponentInst):
        self.instance_struct: NodeStructs.ComponentInst = instance

        self.instance_struct.preannot = instance.ast[0]
        self.instance_JSON = instance.ast[1]
        self.instance_struct.postannot = instance.ast[2]

    def convert(self):
        self._parse()
        self._parse_phases()

        return self.instance_struct

    def _parse_phases(self):
        initSpecs = self.instance_JSON["DefComponentInstance"]["node"]["AstNode"][
            "data"
        ]["initSpecs"]

        for initSpec in initSpecs:
            if "phase" in initSpec[1]["AstNode"]["data"]:
                specType = initSpec[1]["AstNode"]["data"]["phase"]["AstNode"]["data"][
                    "ExprDot"
                ]["id"]["AstNode"]["data"]
                specCode = initSpec[1]["AstNode"]["data"]["code"]

                try:
                    self.instance_struct.phases[specType] = specCode
                except KeyError:
                    print(
                        "[WARN] Phase type not found in instance JSON [InstanceParser.parse_phases]"
                    )

    def _parse(self):
        self.instance_struct.name = self.instance_JSON["DefComponentInstance"]["node"][
            "AstNode"
        ]["data"]["name"]
        self.instance_struct.qf = self.instance_struct.name

        components_bit_1 = self.instance_JSON["DefComponentInstance"]["node"][
            "AstNode"
        ]["data"]["component"]

        self.instance_struct.component_name = Utils.value_parser(components_bit_1)
        self.instance_struct.base_id = Utils.value_parser(
            self.instance_JSON["DefComponentInstance"]["node"]["AstNode"]["data"][
                "baseId"
            ]
        )

        if (
            "queueSize"
            in self.instance_JSON["DefComponentInstance"]["node"]["AstNode"]["data"]
        ):
            self.instance_struct.queue_size = self.instance_JSON[
                "DefComponentInstance"
            ]["node"]["AstNode"]["data"]["queueSize"]
        if (
            "stackSize"
            in self.instance_JSON["DefComponentInstance"]["node"]["AstNode"]["data"]
        ):
            self.instance_struct.stack_size = self.instance_JSON[
                "DefComponentInstance"
            ]["node"]["AstNode"]["data"]["stackSize"]
        if (
            "priority"
            in self.instance_JSON["DefComponentInstance"]["node"]["AstNode"]["data"]
        ):
            self.instance_struct.priority = self.instance_JSON["DefComponentInstance"][
                "node"
            ]["AstNode"]["data"]["priority"]

        if (
            self.instance_struct.queue_size is not None
            and self.instance_struct.queue_size != "None"
        ):
            queueSize_JSON = self.instance_struct.queue_size["Some"]["AstNode"]["data"]

            if "ExprLiteralInt" in queueSize_JSON:
                self.instance_struct.queue_size = queueSize_JSON["ExprLiteralInt"][
                    "value"
                ]
            elif "ExprIdent" in queueSize_JSON:
                self.instance_struct.queue_size = queueSize_JSON["ExprIdent"]["value"]
            elif "ExprDot" in queueSize_JSON:
                self.instance_struct.queue_size = Utils.qualifier_calculator(
                    queueSize_JSON["ExprDot"]
                )

        if (
            self.instance_struct.stack_size is not None
            and self.instance_struct.stack_size != "None"
        ):
            stackSize_JSON = self.instance_struct.stack_size["Some"]["AstNode"]["data"]

            if "ExprLiteralInt" in stackSize_JSON:
                self.instance_struct.stack_size = stackSize_JSON["ExprLiteralInt"][
                    "value"
                ]
            elif "ExprIdent" in stackSize_JSON:
                self.instance_struct.stack_size = stackSize_JSON["ExprIdent"]["value"]
            elif "ExprDot" in stackSize_JSON:
                self.instance_struct.stack_size = Utils.qualifier_calculator(
                    stackSize_JSON["ExprDot"]
                )

        if (
            self.instance_struct.priority is not None
            and self.instance_struct.priority != "None"
        ):
            priority_JSON = self.instance_struct.priority["Some"]["AstNode"]["data"]

            if "ExprLiteralInt" in priority_JSON:
                self.instance_struct.priority = priority_JSON["ExprLiteralInt"]["value"]
            elif "ExprIdent" in priority_JSON:
                self.instance_struct.priority = priority_JSON["ExprIdent"]["value"]
            elif "ExprDot" in priority_JSON:
                self.instance_struct.priority = Utils.qualifier_calculator(
                    priority_JSON["ExprDot"]
                )

        if (
            "cpu"
            in self.instance_JSON["DefComponentInstance"]["node"]["AstNode"]["data"]
        ):
            self.instance_struct.cpu = self.instance_JSON["DefComponentInstance"][
                "node"
            ]["AstNode"]["data"]["cpu"]

            if (
                self.instance_struct.cpu is not None
                and self.instance_struct.cpu != "None"
            ):
                cpu_JSON = self.instance_struct.cpu["Some"]["AstNode"]["data"]

                if "ExprLiteralInt" in cpu_JSON:
                    self.instance_struct.cpu = cpu_JSON["ExprLiteralInt"]["value"]
                elif "ExprIdent" in cpu_JSON:
                    self.instance_struct.cpu = cpu_JSON["ExprIdent"]["value"]
                elif "ExprDot" in cpu_JSON:
                    self.instance_struct.cpu = Utils.qualifier_calculator(
                        cpu_JSON["ExprDot"]
                    )


class PortConverter:
    """
    AST Name: DefPort
    """

    def __init__(self, port: NodeStructs.Port):
        self.port_struct: NodeStructs.Port = port

        self.port_struct.preannot = port.ast[0]
        self.port_JSON = port.ast[1]
        self.port_struct.postannot = port.ast[2]

    def convert(self):
        self.port_struct.name = self.port_JSON["DefPort"]["node"]["AstNode"]["data"][
            "name"
        ]
        self.port_struct.qf = self.port_name

        for i in self.port_JSON["DefPort"]["node"]["AstNode"]["data"]["params"]:
            param = self.port_JSON["DefPort"]["node"]["AstNode"]["data"]["params"][i][1]
            paramToAppend = {
                "name": None,
                "type": None,
            }

            paramToAppend["name"] = param["AstNode"]["data"]["name"]
            paramToAppend["type"] = param["AstNode"]["data"]["typeName"]
            paramToAppend["type"] = Utils.value_parser(paramToAppend["type"])

            self.port_struct.parameters.append(paramToAppend)

        return self.port_struct


class TopologyImportConverter:
    """
    AST Name: SpecTopImport
    """

    def __init__(self, import_struct: NodeStructs.TopologyImport):
        self.import_struct = import_struct

        self.import_struct.preannot = import_struct.ast[0]
        self.import_JSON = import_struct.ast[1]
        self.import_struct.postannot = import_struct.ast[2]

    def convert(self):
        import_bit_1 = self.import_JSON["SpecTopImport"]["node"]["AstNode"]["data"][
            "top"
        ]

        self.import_struct.name = Utils.value_parser(import_bit_1)
        self.import_struct.qf = self.import_struct.name

        return self.import_struct


class ConnectionGraphConverter:
    """
    AST Name: SpecConnectionGraph
    """

    def __init__(self, cg: NodeStructs.ConnectionGraph):
        self.cg_struct = cg

        self.cg_struct.preannot = cg.ast[0]
        self.cg_JSON = cg.ast[1]
        self.cg_struct.postannot = cg.ast[2]

    def convert(self):
        import_bit = self.cg_JSON["SpecConnectionGraph"]["node"]["AstNode"]["data"]

        if "Direct" not in import_bit:
            self.cg_struct.type = "Pattern"
            return
        else:
            self.cg_struct.type = "Direct"
            import_bit = import_bit["Direct"]

        self.cg_struct.name = import_bit["name"]
        self.cg_struct.qf = self.cg_struct.name

        for connection in import_bit["connections"]:
            connectionToAppend = {
                "source": {
                    "name": None,
                    "num": None,
                },
                "dest": {
                    "name": None,
                    "num": None,
                },
            }

            fromPort = connection["fromPort"]["AstNode"]["data"]
            compInst = fromPort["componentInstance"]

            connectionToAppend["source"]["name"] = (
                Utils.value_parser(compInst)
                + "."
                + fromPort["portName"]["AstNode"]["data"]
            )

            if connectionToAppend["source"]["name"][-4:] == "None":
                connectionToAppend["source"]["name"] = connectionToAppend["source"][
                    "name"
                ].replace("None", "")

            if connection["fromIndex"] != "None":
                if connection["fromIndex"] is None or connection["fromIndex"] == "None":
                    connectionToAppend["source"]["num"] = ""
                if type(connection["fromIndex"]) is dict:
                    connectionToAppend["source"]["num"] = (
                        "[" + Utils.value_parser(connection["fromIndex"]["Some"]) + "]"
                    )
                else:
                    connectionToAppend["source"]["num"] = (
                        "[" + connection["fromIndex"] + "]"
                    )

            toPort = connection["toPort"]["AstNode"]["data"]
            compInst = toPort["componentInstance"]

            connectionToAppend["dest"]["name"] = (
                Utils.value_parser(compInst)
                + "."
                + toPort["portName"]["AstNode"]["data"]
            )

            if connectionToAppend["dest"]["name"][-4:] == "None":
                connectionToAppend["dest"]["name"] = connectionToAppend["dest"][
                    "name"
                ].replace("None", "")

            if connection["toIndex"] != "None":
                if connection["toIndex"] is None or connection["toIndex"] == "None":
                    connectionToAppend["dest"]["num"] = ""
                if type(connection["toIndex"]) is dict:
                    connectionToAppend["dest"]["num"] = (
                        "[" + Utils.value_parser(connection["toIndex"]["Some"]) + "]"
                    )
                else:
                    connectionToAppend["dest"]["num"] = (
                        "[" + connection["toIndex"] + "]"
                    )

            self.cg_struct.connections.append(connectionToAppend)

        return self.cg_struct


class TopologyConverter:
    """
    AST Name: DefTopology
    """

    def __init__(self, topology: NodeStructs.Topology):
        self.topology_struct: NodeStructs.Topology = topology

        self.topology_struct.preannot = topology.ast[0]
        self.topology_JSON = topology.ast[1]
        self.topology_struct.postannot = topology.ast[2]

    def convert(self):
        self.topology_struct.name = self.topology_JSON["DefTopology"]["node"][
            "AstNode"
        ]["data"]["name"]
        self.topology_struct.qf = self.topology_struct.name

        self.topology_struct.members = self.topology_JSON["DefTopology"]["node"][
            "AstNode"
        ]["data"]["members"]

        return self.topology_struct



"""fprime.util.file_util: Utility functions for general file operations

@author SiboVG
"""

from pathlib import Path

from fprime.fbuild.builder import Build


def get_directory_path_relative_to_root(build: Build) -> str:
    """
    Get the directory path of the current directory (from which F` tools is executed)
    relative to the root of the F` project.

    If the current directory is the project root, it returns an empty string.
    If the current directory is a subdirectory of the project root, it returns the relative path.
    If the current directory is not within the project root, it returns None.

    Args:
        build (Build): The build object containing project settings.
    Returns:
        str: The relative path from the project root to the current directory,
            an empty string if the current directory is the project root,
            or None if not within the project root.
    """
    proj_root = build.get_settings("project_root", None)
    if proj_root is None:
        print("[WARNING] No project root found. Cannot determine relative path.")
        return None

    proj_root = Path(proj_root).resolve()
    cwd = Path.cwd()

    if proj_root == cwd:
        return ""

    if proj_root not in cwd.parents:
        return None

    return str(cwd.relative_to(proj_root))



"""fprime.fpp.impl: Command line targets for `fprime-util impl`

Processing and CLI entry points for `fprime-util impl` command line tool.

@author thomas-bc
"""

import argparse
import os
import tempfile
from pathlib import Path

from typing import TYPE_CHECKING, Callable, Dict, List, Tuple

from fprime.fbuild.builder import Build

from fprime.fpp.common import FppUtility
from fprime.util.code_formatter import ClangFormatter
from fprime.constants import UT_FILES_TARGET_PATH, UT_TEMPLATE_FILE_SUFFIX


def _apply_clang_formatting(
    build: Build,
    framework_path: Path,
    files_dir: Path,
    generated_file_names: list[Path],
):
    """
    Format files if clang-format is available.

    Args:
        framework_path: F Prime framework path
        files_dir: Folder with generated files
        generated_file_names: List of generated file names
    """

    format_file = framework_path / ".clang-format"
    if format_file.is_file():
        clang_formatter = ClangFormatter("clang-format", format_file, {"backup": False})
        if clang_formatter.is_supported():
            for file_name in generated_file_names:
                clang_formatter.stage_file(files_dir / file_name)
            clang_formatter.execute(build, None, ({}, []))
    else:
        print(
            f"[INFO] .clang-format file not found at {format_file.resolve()}. Skipping formatting."
        )


def _move_ut_templates(files_dir, generated_file_names):
    """
    Move generated UT templates into "files_dir/<UT_FILES_TARGET_PATH>".
    The UT_TEMPLATE_FILE_SUFFIX is added into a file name unless it is already present.

    Args:
        files_dir: Folder with generated files
        generated_file_names: List of generated file names
    """
    # Create the UT folder
    ut_path = files_dir / Path(*UT_FILES_TARGET_PATH)
    ut_path.mkdir(parents=True, exist_ok=True)

    # Move the generated files
    for file_name in generated_file_names:
        source_path = files_dir / file_name
        destination_file_name = (
            file_name.with_suffix(UT_TEMPLATE_FILE_SUFFIX + file_name.suffix)
            if UT_TEMPLATE_FILE_SUFFIX not in file_name.suffixes
            else file_name
        )
        destination_path = ut_path / destination_file_name
        source_path.rename(destination_path)


def fpp_generate_implementation(
    build: Build,
    output_dir: Path,
    context: Path,
    apply_formatting: bool,
    generate_ut: bool,
    generate_test_helpers: bool = False,
) -> int:
    """
    Generate implementation files from FPP templates.

    Args:
        build: Build object
        output_dir: The directory where the generated files will be written
        context: The path to the F´ module to generate files for
        apply_formatting: Whether to format the generated files using clang-format
        generate_ut: Generates UT files if set to True
        generate_test_helpers: Generate of test helper code if set to True
    """

    prefixes = [
        build.get_settings("framework_path", ""),
        *build.get_settings("library_locations", []),
        build.get_settings("project_root", ""),
        build.build_dir / "F-Prime",
        build.build_dir,
    ]

    # Holds the list of generated files to be passed to clang-format
    gen_files = tempfile.NamedTemporaryFile(prefix="fprime-impl-")

    output_dir.mkdir(parents=True, exist_ok=True)

    # Run fpp-to-cpp --template
    FppUtility("fpp-to-cpp", imports_as_sources=False).execute(
        build,
        context,
        args=(
            {},
            [
                "--template",
                *(["--unit-test"] if generate_ut else []),
                *(["--auto-test-helpers"] if not generate_test_helpers else []),
                "--names",
                gen_files.name,
                "--directory",
                str(output_dir),
                "--path-prefixes",
                ",".join(map(str, prefixes)),
            ],
        ),
    )

    framework_path = build.settings.get("framework_path", Path("."))
    # FPP --names outputs a list of file names.
    generated_file_names = [
        Path(line.decode("utf-8").strip()) for line in gen_files.readlines()
    ]

    if apply_formatting:
        _apply_clang_formatting(build, framework_path, output_dir, generated_file_names)

    if generate_ut:
        _move_ut_templates(output_dir, generated_file_names)

    return 0


def run_fpp_impl(
    build: Build,
    parsed: argparse.Namespace,
    _: Dict[str, str],
    __: Dict[str, str],
    ___: List[str],
):
    """

    Args:
        build: build object
        parsed: parsed input arguments
        _: unused cmake_args
        __: unused make_args
        ___: unused pass-through arguments
    """

    return fpp_generate_implementation(
        build,
        Path(parsed.output_dir),
        Path(parsed.path),
        not parsed.no_format,
        parsed.ut,
        parsed.generate_test_helpers,
    )


def add_fpp_impl_parsers(
    subparsers, common: argparse.ArgumentParser
) -> Tuple[Dict[str, Callable], Dict[str, argparse.ArgumentParser]]:
    """Sets up the fprime-viz command line parsers

    Creates command line parsers for fprime-viz commands and associates these commands to processing
    functions for those fpp commands.

    Args:
        subparsers: subparsers to add to
        common: common parser for all fprime-util commands

    Returns:
        Tuple of dictionary mapping command name to processor, and command to parser
    """
    impl_parser = subparsers.add_parser(
        "impl",
        help="Generate implementation templates",
        parents=[common],
        add_help=False,
    )
    impl_parser.add_argument(
        "--output-dir",
        help="Directory to generate files in. Default: cwd",
        required=False,
        default=os.getcwd(),
    )
    impl_parser.add_argument(
        "--no-format",
        action="store_true",
        help="Disable formatting (using clang-format) of generated files",
        required=False,
    )
    impl_parser.add_argument(
        "--generate-test-helpers",
        action="store_true",
        default=False,
        help="Generate test helper code for hand-coding. Default to False, leveraging the test helpers autocoded by FPP.",
        required=False,
    )
    return {"impl": run_fpp_impl}, {"impl": impl_parser}



"""fprime.fbuild.cli: fbuild command line processing

Command line processing functions for fbuild (fprime build system targets) of "generate", "purge", and build system
target operations.

@author mstarch
"""

import argparse
from pathlib import Path
from typing import Callable, Dict, List, Tuple

from fprime.common.utils import confirm
from fprime.fbuild.builder import Build
from fprime.fbuild.enumerator import DesignateTargetAction
from fprime.fbuild.target import Target
from fprime.fbuild.types import BuildType


def get_target(parsed: argparse.Namespace) -> Target:
    """Gets the target given the argparse namespace

    Takes the parsed namespace and processes it to a known matching target.

    Args:
        parsed: argparse namespace to read values from

    Returns:
        target that can support the given supplied namespace
    """
    mnemonic = parsed.command
    flags = {
        flag for flag in Target.get_all_possible_flags() if getattr(parsed, flag, False)
    }
    return Target.get_target(mnemonic, flags)


def run_fbuild_cli(
    build: Build,
    parsed: argparse.Namespace,
    cmake_args: Dict[str, str],
    make_args: Dict[str, str],
    pass_through: List[str],
):
    """Execution of the fbuild commands

    Args:
        build: build object, pre-loaded, iff not "generate" or "purged"
        parsed: parsed arguments object.
        cmake_args: cmake commands from parsed output
        make_args: arguments to supply to the build tool (make or ninja)
    """
    if parsed.command == "generate":
        if parsed.force and build.build_dir.exists():
            print(f"[INFO] Purging build directory at: {build.build_dir}")
            build.purge()

        toolchain = build.find_toolchain()
        print(f"[INFO] Generating build directory at: {build.build_dir}")
        print(f"[INFO] Using toolchain file {toolchain} for platform {parsed.platform}")
        if parsed.disable_sanitizers:
            # The following options are defined in F' to have CMake enable the sanitizers
            cmake_args["ENABLE_SANITIZER_LEAK"] = "OFF"
            cmake_args["ENABLE_SANITIZER_ADDRESS"] = "OFF"
            cmake_args["ENABLE_SANITIZER_UNDEFINED_BEHAVIOR"] = "OFF"
        if toolchain is not None:
            cmake_args["CMAKE_TOOLCHAIN_FILE"] = toolchain

        if parsed.make:
            # No need to change CMAKE_GENERATOR since the default is Unix Makefiles
            pass
        else:
            # Sets Ninja as the default build system
            cmake_args["CMAKE_GENERATOR"] = "Ninja"

        build.generate(cmake_args)
    elif parsed.command == "purge":
        # Since purge does not load its "base", we need to overload the platform
        purge(build, parsed)

    else:
        target = get_target(parsed)
        option_args = {
            flag: getattr(parsed, flag.lstrip("-").replace("-", "_"), False)
            for flag, _ in target.option_args()
        }
        target.execute(
            build,
            context=Path(parsed.path),
            args=(make_args, pass_through, option_args),
        )


def purge(build: Build, parsed: argparse.Namespace):
    for purge_build in Build.get_build_list(
        build, parsed.build_cache, ignore_invalid=parsed.force
    ):
        print(
            f"[INFO] {parsed.command.title()} build directory at: {purge_build.build_dir}"
        )
        try:
            perform_purge = False
            if parsed.force:
                if Path(purge_build.build_dir).is_dir():
                    perform_purge = True
                else:
                    print(
                        f"[INFO] Skipping purge. The following directory does not exist: {purge_build.build_dir}"
                    )
            else:
                perform_purge = confirm("Purge this directory?")

            if perform_purge:
                purge_build.purge()

            install_dir = purge_build.install_dest_exists()
            if (
                purge_build.build_type != BuildType.BUILD_CUSTOM
                and install_dir
                and install_dir.exists()
            ):
                print(
                    f"[INFO] {parsed.command.title()} install directory at: {install_dir}"
                )
                if parsed.force or confirm(
                    f"Purge installation directory at {install_dir} ?"
                ):
                    purge_build.purge_install()
        except PermissionError as e:
            print(
                f"Error: Permission denied while purging {purge_build.build_dir}: {e}"
            )


def add_target_parser(
    target: Target,
    subparsers,
    common: argparse.ArgumentParser,
    existing: Dict[str, Tuple[argparse.ArgumentParser, List[str]]],
    help_text: "HelpText",
) -> str:
    """Add a subparser for a given build target

    For a given build target, construct an argument parser with mnemonic as subcommand and needed flags. Then add it as
    a subparser to the global parent. If it already exists in existing, then just add non-existent flags/

    Args:
        target:     target for building a new subparser
        subparsers: argument parser to add a subparser to
        common:     common subparser to be used as parent carrying common flags
        existing:   dictionary storing the mnemonic to parser and flags tuple
        help_text:  mnemonic to help text mapping

    Returns:
        Name of the command this parser handles

    Notes:
        This functions has side effects of editing existing and the list of subparsers
    """
    help_string = help_text.short(
        target.mnemonic, f"{target.desc} in the specified directory"
    )
    description = help_text.long(target.mnemonic, help_string)
    if target.mnemonic not in existing:
        parser = subparsers.add_parser(
            target.mnemonic,
            parents=[common],
            add_help=False,
            description=description,
            help=help_string,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        # --ut flag also exists at the global parsers, skip adding it
        existing[target.mnemonic] = (parser, ["ut"])
        # Common target-only items
        parser.add_argument(
            "-j",
            "--jobs",
            type=int,
            help="Parallel build job count. Default: build tool default (make: 1, ninja: # of cores)",
        )
    parser, flags = existing[target.mnemonic]
    new_flags = [flag for flag in target.flags if flag not in flags]
    for flag in new_flags:
        extra_help = (
            f". Use '--pass-through' to supply '{target.pass_handler()}' arguments"
            if target.allows_pass_args()
            else ""
        )
        # Target flag special handling
        if flag == "target":
            parser.add_argument(
                f"--target",
                type=str,
                action=DesignateTargetAction,
                default=None,
                nargs=1,
                help=f"{target.desc}",
            )
        else:
            parser.add_argument(
                f"--{flag}",
                action="store_true",
                default=False,
                help=f"{target.desc}{extra_help}",
            )
    for flag, description in filter(
        lambda flag_desc: flag_desc[0] not in flags, target.option_args()
    ):
        parser.add_argument(flag, action="store_true", help=f"{description}")
        new_flags.append(flag)
    # Allow pass through arguments
    if target.pass_handler() and "--pass-through" not in flags:
        parser.add_argument(
            "--pass-through",
            nargs=argparse.REMAINDER,
            default=[],
            help="If specified, --pass-through must be the last argument. Remaining arguments passed to underlying executable",
        )
        flags.append("--pass-through")
    flags.extend(new_flags)
    return target.mnemonic


def add_special_targets(
    subparsers, common: argparse.ArgumentParser, help_text: "HelpText"
) -> List[str]:
    """Add in generate and purge commands

    Args:
        subparsers: subparsers used to create new subparsers
        common: common parser
        help_text: map of mnemonic to help text

    Returns:
        ["generate", "purge"] list of special targets
    """

    generate_parser = subparsers.add_parser(
        "generate",
        help=help_text.short(
            "generate", "Generate a build cache for specified project"
        ),
        description=help_text.long(
            "generate", "Generate a build cache for specified project"
        ),
        parents=[common],
        add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    generate_parser.add_argument(
        "--disable-sanitizers",
        default=False,
        help="Disable the compiler sanitizers. Sanitizers are only enabled by default when --ut is provided.",
        action="store_true",
    )
    generate_parser.add_argument(
        "--make",
        action="store_true",
        help="Use the Unix Makefiles build system instead of Ninja",
    )
    # The following option is specified only to show up in --help.
    # It is not handled by argparse, but in fprime.util.cli:validate()
    generate_parser.add_argument(
        "-D<VAR>[:<TYPE>]=<VALUE>",
        action="store_true",
        help="Pass -D flags through to CMake. Can be used multiple times.",
    )
    generate_parser.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Before generating, purges the build directory by force if it already exists. No confirmation will be requested.",
    )
    purge_parser = subparsers.add_parser(
        "purge",
        help=help_text.short("purge", "Remove build caches for specified project"),
        description=help_text.long(
            "purge", "Remove build caches for specified project"
        ),
        add_help=False,
        parents=[common],
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    purge_parser.add_argument(
        "-f",
        "--force",
        default=False,
        action="store_true",
        help="Purges the build directory by force. No confirmation will be requested.",
    )
    return ["generate", "purge"]


def add_fbuild_parsers(
    subparsers, common: argparse.ArgumentParser, help_text: "HelpText"
) -> Tuple[Dict[str, Callable], Dict[str, argparse.ArgumentParser]]:
    """Add in the build-system targets: generate, purge, and all build endpoints

    Sets up the targets used to generate and run the build system. This includes the purge command for cleaning up. This
    wraps up all targets defined in the targets list in fprime.fbuild.builder

    Args:
        subparsers: subparsers object used to add subparsers to the global CLI
        common: common parser used to add all common arguments
        help_text: map of mnemonics to help text

    Returns:
        Tuple of two dictionaries mapping command name to the function used to process commands and to parser
    """
    parsers = {}
    run_map = {
        add_target_parser(
            target, subparsers, common, parsers, help_text=help_text
        ): run_fbuild_cli
        for target in Target.get_all_targets()
    }
    run_map.update(
        {
            command: run_fbuild_cli
            for command in add_special_targets(subparsers, common, help_text=help_text)
        }
    )
    return run_map, parsers



"""
Created on Dec 18, 2014
@author: tcanham, reder
"""

import struct

from .type_base import DictionaryType
from .type_exceptions import (
    DeserializeException,
    EnumMismatchException,
    NotInitializedException,
    TypeMismatchException,
    TypeRangeException,
    InvalidRepresentationTypeException,
    RepresentationTypeRangeException,
)
from .numerical_types import IntegerType

REPRESENTATION_TYPE_MAP = {
    cls.get_canonical_name(): cls for cls in IntegerType.__subclasses__()
}


class EnumType(DictionaryType):
    """
    Representation of the ENUM type.

    The enumeration takes a dictionary that stores the enumeration members
    and current value as a string. The member values will have to be computed
    containing code based on C enum rules
    """

    @classmethod
    def construct_type(cls, name, enum_dict, rep_type="I32"):
        """Construct the custom enum type

        Constructs the custom enumeration type, with the supplied enumeration dictionary.

        Args:
            name: name of the enumeration type
            enum_dict: enumeration: value dictionary defining the enumeration
            rep_type: representation type (standard Fprime integer types)
        """
        if not isinstance(enum_dict, dict):
            raise TypeMismatchException(dict, type(enum_dict))
        for member in enum_dict.keys():
            if not isinstance(member, str):
                raise TypeMismatchException(str, type(member))
            if not isinstance(enum_dict[member], int):
                raise TypeMismatchException(int, enum_dict[member])

        if rep_type not in REPRESENTATION_TYPE_MAP.keys():
            raise InvalidRepresentationTypeException(
                rep_type, REPRESENTATION_TYPE_MAP.keys()
            )

        for member in enum_dict.keys():
            type_range = REPRESENTATION_TYPE_MAP[rep_type].range()
            if enum_dict[member] < type_range[0] or enum_dict[member] > type_range[1]:
                raise RepresentationTypeRangeException(
                    member, enum_dict[member], rep_type, type_range
                )

        return DictionaryType.construct_type(
            cls, name, ENUM_DICT=enum_dict, REP_TYPE=rep_type
        )

    @classmethod
    def validate(cls, val):
        """Validate the value passed into the enumeration"""
        if not isinstance(val, str):
            raise TypeMismatchException(str, type(val))
        if val not in cls.keys():
            raise EnumMismatchException(cls.__class__.__name__, val)

    @classmethod
    def keys(cls):
        """
        Return all the enum key values.
        """
        return list(cls.ENUM_DICT.keys())

    def serialize(self):
        """
        Serialize the enumeration type using an int type
        """
        # for enums, take the string value and convert it to
        # the numeric equivalent
        if self._val is None or (
            self._val == "UNDEFINED" and "UNDEFINED" not in self.ENUM_DICT
        ):
            raise NotInitializedException(type(self))
        return struct.pack(
            REPRESENTATION_TYPE_MAP[self.REP_TYPE].get_serialize_format(),
            self.ENUM_DICT[self._val],
        )

    def deserialize(self, data, offset):
        """
        Deserialize the enumeration using an int type
        """
        try:
            int_val = struct.unpack_from(
                REPRESENTATION_TYPE_MAP[self.REP_TYPE].get_serialize_format(),
                data,
                offset,
            )[0]

        except struct.error:
            msg = f"Could not deserialize enum value. Needed: {self.getSize()} bytes Found: {len(data[offset:])}"
            raise DeserializeException(msg)
        for key, val in self.ENUM_DICT.items():
            if int_val == val:
                self._val = key
                break
        # Value not found, invalid enumeration value
        else:
            raise TypeRangeException(int_val)

    def getSize(self):
        """Calculates the size based on the size of an integer used to store it"""
        return struct.calcsize(
            REPRESENTATION_TYPE_MAP[self.REP_TYPE].get_serialize_format()
        )

    @classmethod
    def getMaxSize(cls):
        """Maximum size of type"""
        return struct.calcsize(
            REPRESENTATION_TYPE_MAP[cls.REP_TYPE].get_serialize_format()
        )



from pathlib import Path
from fprime.fpp.utils.fpp_to_json.helpers import openFppFile
import fprime.fpp.utils.fpp_to_json.visitors.json_conversion as JSONConverter
import fprime.fpp.utils.fpp_to_json.node_structs as NodeStructs


def walkModule(data, oldQf):
    module = NodeStructs.Module(data)
    module: NodeStructs.Module = JSONConverter.ModuleConverter(module).convert()

    if oldQf == "":
        qf = module.name
    else:
        qf = oldQf + "." + module.name  # qualifier

    module.qf = qf

    for member in module.members:
        if "DefComponentInstance" in member[1]:
            instance = NodeStructs.ComponentInst(member)
            instance: NodeStructs.ComponentInst = JSONConverter.CompInstanceConverter(
                instance
            ).convert()
            instance.qf = qf + "." + instance.name

        if "DefConstant" in member[1]:
            constant = NodeStructs.Constant(member)
            constant: NodeStructs.Constant = JSONConverter.ConstantConverter(
                constant
            ).convert()
            constant.qf = qf + "." + constant.id

        # can be continued for other member types

        if "DefTopology" in member[1]:
            walkTopology(member, qf)

        if "DefModule" in member[1]:
            walkModule(member, qf)

    return qf


def walkTopology(data, module):
    topology = NodeStructs.Topology(data)
    topology: NodeStructs.Topology = JSONConverter.TopologyConverter(topology).convert()

    if module == "":
        qf = topology.name
    else:
        qf = module + "." + topology.name  # qualifier

    topology.qf = qf

    for member in topology.members:
        if "DefTopology" in member[1]:
            walkTopology(member, qf)

        if "DefModule" in member[1]:
            walkModule(member, qf)

    return qf


def visit():
    AST = openFppFile(Path(__file__).parent / "example.fpp")

    for i in range(len(AST[0]["members"])):
        if "DefModule" in AST[0]["members"][i][1]:
            walkModule(AST[0]["members"][i], "")

        if "DefTopology" in AST[0]["members"][i][1]:
            walkTopology(AST[0]["members"][i], "")


if __name__ == "__main__":
    visit()



"""
(test) fprime-utils commands:

Tests the F prime util commands.
@author 0x48piraj
"""

import unittest


class CommandsTestCases(unittest.TestCase):
    def test_build_command(self):
        pass

    def test_check_command(self):
        pass

    def test_generate_command(self):
        pass

    def test_purge_command(self):
        pass

    def test_fpp_check_command(self):
        pass

    def test_visualize_command(self):
        pass

    def test_impl_command(self):
        pass

    def test_hash_to_file_command(self):
        pass

    def test_info_command(self):
        pass

    def test_new_command(self):
        pass

    def test_format_command(self):
        pass

    def test_version_check_command(self):
        pass


if __name__ == "__main__":
    unittest.main()



"""
(test) fprime.fbuild.settings:

Tests the F prime settings module.
@author joshuaa
"""

import os
from pathlib import Path

from fprime.fbuild.settings import IniSettings

LOCAL_PATH = Path(__file__).parent


def full_path(path):
    path = LOCAL_PATH / Path(path)
    return path.resolve()


# The following tests use a fake framework path due to the separation of fprime and fprime-tools


def test_settings():
    test_cases = [
        {
            "file": "settings-empty.ini",
            "expected": {
                "settings_file": full_path("settings-data/settings-empty.ini"),
                "default_toolchain": "native",
                "default_ut_toolchain": "native",
                "framework_path": full_path(".."),
                "install_destination": full_path("settings-data/build-artifacts"),
                "library_locations": [],
                "environment_file": full_path("settings-data/settings-empty.ini"),
                "environment": {},
                "component_cookiecutter": "default",
                "deployment_cookiecutter": "default",
                "project_root": full_path(".."),
                "config_directory": full_path("..") / "config",
                "default_cmake_options": "",
            },
        },
        {
            "file": "settings-custom-install.ini",
            "expected": {
                "settings_file": full_path("settings-data/settings-custom-install.ini"),
                "default_toolchain": "native",
                "default_ut_toolchain": "native",
                "framework_path": full_path(".."),
                "install_destination": full_path("test"),
                "library_locations": [],
                "environment_file": full_path(
                    "settings-data/settings-custom-install.ini"
                ),
                "environment": {},
                "component_cookiecutter": "default",
                "deployment_cookiecutter": "default",
                "project_root": full_path(".."),
                "config_directory": full_path("..") / "config",
                "default_cmake_options": "",
            },
        },
        {
            "file": "settings-custom-toolchain.ini",
            "expected": {
                "settings_file": full_path(
                    "settings-data/settings-custom-toolchain.ini"
                ),
                "default_toolchain": "custom1",
                "default_ut_toolchain": "custom2",
                "framework_path": full_path(".."),
                "install_destination": full_path("settings-data/build-artifacts"),
                "library_locations": [],
                "environment_file": full_path(
                    "settings-data/settings-custom-toolchain.ini"
                ),
                "environment": {},
                "component_cookiecutter": "default",
                "deployment_cookiecutter": "default",
                "project_root": full_path(".."),
                "config_directory": full_path("..") / "config",
                "default_cmake_options": "",
            },
        },
        {
            "file": "settings-outside-cookiecutter.ini",
            "expected": {
                "settings_file": full_path(
                    "settings-data/settings-outside-cookiecutter.ini"
                ),
                "default_toolchain": "native",
                "default_ut_toolchain": "native",
                "framework_path": full_path(".."),
                "install_destination": full_path("settings-data/build-artifacts"),
                "library_locations": [],
                "environment_file": full_path(
                    "settings-data/settings-outside-cookiecutter.ini"
                ),
                "environment": {},
                "component_cookiecutter": "gh:SterlingPeet/cookiecutter-fprime-deployment",
                "deployment_cookiecutter": "https://github.com/thomas-bc/fprime-deployment-cookiecutter.git",
                "project_root": full_path(".."),
                "config_directory": full_path("..") / "config",
                "default_cmake_options": "",
            },
        },
        {
            "file": "settings-multi-line-default-options.ini",
            "expected": {
                "settings_file": full_path(
                    "settings-data/settings-multi-line-default-options.ini"
                ),
                "default_toolchain": "native",
                "default_ut_toolchain": "native",
                "framework_path": full_path(".."),
                "install_destination": full_path("settings-data/build-artifacts"),
                "library_locations": [],
                "environment_file": full_path(
                    "settings-data/settings-multi-line-default-options.ini"
                ),
                "environment": {},
                "component_cookiecutter": "default",
                "deployment_cookiecutter": "default",
                "project_root": full_path(".."),
                "config_directory": full_path("..") / "config",
                "default_cmake_options": "OPTION1=ABC\nOPTION2=123\nOPTION3=Something",
            },
        },
        {
            "file": "settings-environment.ini",
            "expected": {
                "settings_file": full_path("settings-data/settings-environment.ini"),
                "default_toolchain": "native",
                "default_ut_toolchain": "native",
                "framework_path": full_path(".."),
                "install_destination": full_path("settings-data/build-artifacts"),
                "library_locations": [],
                "environment_file": full_path("settings-data/settings-environment.ini"),
                "environment": {"MY_VARIABLE": "my value", "MY_VARIABLE_2": "abc:123"},
                "component_cookiecutter": "default",
                "deployment_cookiecutter": "default",
                "project_root": full_path(".."),
                "config_directory": full_path("..") / "config",
                "default_cmake_options": "",
            },
        },
    ]
    # Prep for substitution
    os.environ["TEST_SETTING_1"] = "abc"
    os.environ["TEST_SETTING_2"] = "123"
    for case in test_cases:
        fp = full_path("settings-data/" + case["file"])
        results = IniSettings.load(fp)
        assert (
            case["expected"] == results
        ), f'{fp}: Expected {case["expected"]}, got {results}'



import subprocess
import os


def fpp_depend(cache_folder, input_file, locs_files) -> str:
    """
    This function calculates the dependencies for an fpp file using fprime-util to get
    the location of the build cache fpp-depend.

    Args:
        input_file: The input fpp file to calculate dependencies for
        locs_file: The locs.fpp file to use for dependency calculation

    Returns:
        A string of dependencies for the input file
    """

    print(f"[fpp] Calculating fpp dependencies for {os.path.basename(input_file)}...")

    try:
        fppDep = subprocess.run(
            ["fpp-depend", input_file]
            + locs_files
            + [
                "-d",
                f"{cache_folder}/direct.txt",
                "-m",
                f"{cache_folder}/missing.txt",
                "-f",
                f"{cache_folder}/framework.txt",
                "-g",
                f"{cache_folder}/generated.txt",
                "-i",
                f"{cache_folder}/include.txt",
                "-u",
                f"{cache_folder}/unittest.txt",
                "-a",
            ],
            check=True,
            stdout=subprocess.PIPE,
        )

        with open(f"{cache_folder}/stdout.txt", "w") as f:
            f.write(fppDep.stdout.decode("utf-8"))

    except subprocess.CalledProcessError as e:
        print(f"[ERR] fpp-depend failed with error: {e}")
        return 1

    raise Exception("fpp-depend failed")


def compute_simple_dependencies(locs_file, input):
    print(f"[fpp] Calculating simple fpp dependencies for {os.path.basename(input)}...")

    try:
        fppDep = subprocess.run(
            ["fpp-depend", locs_file, input],
            check=True,
            stdout=subprocess.PIPE,
        )

        return fppDep.stdout.decode("utf-8")
    except subprocess.CalledProcessError as e:
        print(f"[ERR] fpp-depend failed with error: {e}")
        return 1


def fpp_to_json(input_file):
    """
    This function runs fpp-to-json on an fpp file to generate a JSON AST.

    Args:
        input_file: The input fpp file to run fpp-to-json on

    Returns:
        None
    """

    # run fpp
    print(f"[fpp] Running fpp-to-json for {os.path.basename(input_file)}...")

    cmdS = ["fpp-to-json", input_file, "-s"]

    try:
        subprocess.run(cmdS, check=True, stdout=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        raise Exception(f"[ERR] fpp-to-json pt2 failed with error: {e}")


def fpp_format(input_file):
    """
    This function runs fpp-format on an fpp file to format the file.

    Args:
        input_file: The input fpp file to run fpp-format on

    Returns:
        None
    """

    # run fpp-format
    print(f"[fpp] Running fpp-format for {os.path.basename(input_file)}...")

    try:
        fppFormat = subprocess.run(
            ["fpp-format", input_file], check=True, stdout=subprocess.PIPE
        )
        return fppFormat.stdout.decode("utf-8")
    except subprocess.CalledProcessError as e:
        print(f"[ERR] fpp-format failed with error: {e}")
        return 1


def fpp_locate_defs(input_file):
    """
    This function runs fpp-locate-defs on an fpp file to locate definitions.

    Args:
        input_file: The input fpp file to run fpp-locate-defs on
        locs_file:  The locs.fpp file used to find the base directory to base def locations
                    off of
    """

    print(f"[fpp] Running fpp-locate-defs for {os.path.basename(input_file)}...")

    base_dir = os.path.dirname(input_file)

    try:
        fppLocateDefs = subprocess.run(
            ["fpp-locate-defs", input_file, "-d", base_dir],
            check=True,
            stdout=subprocess.PIPE,
        )
        return fppLocateDefs.stdout.decode("utf-8")
    except subprocess.CalledProcessError as e:
        print(f"[ERR] fpp-locate-defs failed with error: {e}")
        return 1



#!/usr/bin/env python
"""Fake CMake script for testing purposes

This script can be added to the front of the path to echo what arguments are sent to CMake allowing for tests that are
slightly closer to integration tests without involving the entirety of F Prime.
"""
import sys
import os
from pathlib import Path
import json

if __name__ == "__main__":
    print("[INFO] Running echoer program (stdout)")
    print("[INFO] Running echoer program (stderr)", file=sys.stderr)
    executable_path = Path(sys.argv[0])
    for i in range(0, 100):
        output_file_path = (
            executable_path.parent / f"faux-{executable_path.name}-{i}.json"
        )
        if not output_file_path.exists():
            with open(output_file_path, "w") as output_file:
                json.dump(
                    {
                        "arguments": sys.argv,
                        "cwd": str(Path.cwd()),
                        "environment": dict(os.environ),
                    },
                    output_file,
                )
                break
    else:
        print(
            f"[ERROR] Too many invocations of: {executable_path.name}", file=sys.stderr
        )
        sys.exit(1)
    sys.exit(0)



"""
fprime.util.build_helper.py

This script is defined to help users run standard make processes using CMake. This will support migrants from the old
make system, as well as enable more efficient practices when developing in the CMake system. The following functions
are supported herein:

 - build: build the current directory/module/deployment
 - impl: make implementation templates
 - testimpl: make testing templates
 - build_ut: build the current UTs
 - check: run modules unit tests

@author mstarch
"""

from pathlib import Path

from fprime.fbuild.builder import Build, BuildType
from fprime.fbuild.cli import get_target
from fprime.fbuild.target import NoSuchTargetException

from .versioning import VersionException, get_version, FPRIME_PIP_PACKAGES

# Attempt to get pkg_resources from "setuptools"
try:
    import pkg_resources
except ImportError:
    pkg_resources = None


def package_version_check(package: str, requirement_path: Path):
    """Checks the version of the packages installed match the expected packages of the fprime aggregate package"""
    expected_version = get_version(package, requirement_path).lstrip(
        "v"
    )  # Python version
    try:
        version = pkg_resources.get_distribution(package).version
        if version != expected_version:
            print(
                f"[WARNING] {package} has unexpected version. Expected: {expected_version} found {version}"
            )
    except pkg_resources.DistributionNotFound:
        print(f"[WARNING] {package} is not installed")


def validate_tools_from_requirements(build: Build):
    """Uses build settings to find requirements file and validate the correct versions installed"""
    # Find prioritized order of requirements.txt, ensuring that each member exists
    possibilities = [
        build.settings.get("project_root", None),
        build.settings.get("framework_path", None),
    ]
    possibilities = [
        Path(possible) / "requirements.txt"
        for possible in possibilities
        if possible is not None
    ]
    possibilities = [possible for possible in possibilities if possible.exists()]
    # Skip tools check, as not requirements.txt found
    if not possibilities:
        print(
            f"[WARNING] Could not find 'requirements.txt' in: {possibilities}. Will not check tool versions."
        )
        return
    # Pre-roll import errors from pkg_resources
    if pkg_resources is None:
        print("[WARNING] Cannot import 'pkg_resources'. Will not check tool versions.")
        return

    # Now check each required tool for fprime
    for tool in FPRIME_PIP_PACKAGES:
        for possible in possibilities:
            try:
                package_version_check(tool, possible)
                break
            except (OSError, VersionException) as exc:
                message = f"[WARNING] {exc}"
        else:
            print(message)


def load_build(parsed, skip_validation=False):
    """
    Loads Build object and returns it to the caller. Additionally, this will validate the
    installed tool versions (such as fpp and other F' utilities) against the requirements.txt
    """

    try:
        target = get_target(parsed)
        build_type = target.build_type
    except NoSuchTargetException:
        build_type = BuildType.BUILD_TESTING if parsed.ut else BuildType.BUILD_NORMAL

    cmake_root = (
        Path(parsed.root)
        if parsed.root is not None
        else Build.find_nearest_parent_project(Path.cwd())
    )

    build = Build(build_type, cmake_root, verbose=parsed.verbose)

    # All commands need to load the build cache to setup the basic information for the build with the exception of
    # generate, which is run before the creation of the build cache and thus must invent the cache instead. This
    # call will ensure the build is in a ready state before attempting to check tool versions and run the command.
    #
    # Some commands, like purge and info, run on sets of directories and will attempt to load those sets later.
    # However, the base directory must be setup here. Errors in this load are ignored to allow the command to find
    # build caches related to that set.
    if parsed.command == "generate":
        build.invent(parsed.platform, build_dir=parsed.build_cache, force=parsed.force)
    else:
        build.load(
            parsed.platform,
            parsed.build_cache,
            skip_validation=skip_validation,
        )
    validate_tools_from_requirements(build)
    return build



"""
Tests for fprime.fbuild.target
"""

from pathlib import Path
from unittest.mock import MagicMock

import pytest

from fprime.fbuild.target import (
    Target,
    TargetScope,
    CompositeTarget,
    BuildSystemTarget,
    NoSuchTargetException,
    EnumeratedAction,
)
from fprime.fbuild.enumerator import BuildTargetEnumerator


@pytest.fixture(autouse=True)
def clear_registered_targets():
    """Fixture to clear registered targets before each test"""
    Target._Target__REGISTRY = {}
    yield
    Target._Target__REGISTRY = {}


@pytest.fixture
def mock_builder():
    """Pytest fixture for a mock builder object"""
    builder = MagicMock()
    builder.is_verbose.return_value = False
    return builder


def test_register_and_get_target():
    """Test target registration and retrieval"""
    target = Target("test", "A test target", TargetScope.LOCAL)
    Target.register_target(target)

    found_target = Target.get_target("test", set())
    assert found_target == target

    with pytest.raises(NoSuchTargetException):
        Target.get_target("nonexistent", set())


def test_composite_target_execution(mock_builder):
    """Test that a composite target executes its children"""
    child1 = MagicMock(spec=Target, scope=TargetScope.LOCAL)
    child2 = MagicMock(spec=Target, scope=TargetScope.LOCAL)
    composite = CompositeTarget([child1, child2], "composite", "", TargetScope.LOCAL)

    composite.execute(mock_builder, Path("."), (None, None, None))

    child1.execute.assert_called_once()
    child2.execute.assert_called_once()


def test_build_system_target_execution(mock_builder):
    """Test that BuildSystemTarget calls the build system"""
    target = BuildSystemTarget("build", "", TargetScope.LOCAL)
    target.original_context = Path(".")

    target.execute_one(mock_builder, "target_name", ({}, [], {}))

    mock_builder.execute_build_target.assert_called_once_with(
        "target_name", Path("."), {}
    )


def test_enumerated_action_execution(mock_builder):
    """Test that EnumeratedAction executes its enumerator and then execute_all"""
    mock_enumerator = MagicMock(spec=BuildTargetEnumerator)
    enumerated_results = ["target1", "target2"]
    mock_enumerator.enumerate.return_value = enumerated_results

    class ConcreteEnumeratedAction(EnumeratedAction):
        def execute_all(self, builder, context, args):
            pass  # This will be mocked

        def any_supported(self, builder, context):
            return True

    action = ConcreteEnumeratedAction(TargetScope.LOCAL, mock_enumerator)
    action.execute_all = MagicMock()

    context_path = Path("/path/to/context")
    args = ({}, [], {})
    action.execute(mock_builder, context_path, args)

    mock_enumerator.enumerate.assert_called_once_with(mock_builder, context_path)
    action.execute_all.assert_called_once_with(mock_builder, enumerated_results, args)


def test_enumerated_action_is_supported(mock_builder):
    """Test that EnumeratedAction.is_supported calls the enumerator and any_supported"""
    mock_enumerator = MagicMock(spec=BuildTargetEnumerator)
    enumerated_results = ["target1", "target2"]
    mock_enumerator.enumerate.return_value = enumerated_results

    class ConcreteEnumeratedAction(EnumeratedAction):
        def execute_all(self, builder, context, args):
            pass

        def any_supported(self, builder, context):
            return True  # This will be mocked

    action = ConcreteEnumeratedAction(TargetScope.LOCAL, mock_enumerator)
    action.any_supported = MagicMock(return_value=True)

    context_path = Path("/path/to/context")
    result = action.is_supported(mock_builder, context_path)

    assert result is True
    mock_enumerator.enumerate.assert_called_once_with(mock_builder, context_path)
    action.any_supported.assert_called_once_with(mock_builder, enumerated_results)


@pytest.mark.parametrize(
    "child_supported, expected_composite_supported",
    [
        ([True, True], True),
        ([True, False], False),
        ([False, False], False),
    ],
)
def test_composite_target_is_supported(
    child_supported, expected_composite_supported, mock_builder
):
    """Test CompositeTarget.is_supported logic"""
    child1 = MagicMock(spec=Target)
    child1.is_supported.return_value = child_supported[0]
    child2 = MagicMock(spec=Target)
    child2.is_supported.return_value = child_supported[1]

    composite = CompositeTarget([child1, child2], "composite", "", TargetScope.LOCAL)
    result = composite.is_supported(mock_builder, Path("."))

    assert result == expected_composite_supported


def test_composite_target_option_args():
    """Test CompositeTarget.option_args aggregation and deduplication"""
    child1 = MagicMock(spec=Target)
    child1.option_args.return_value = [("--opt1", "Help 1"), ("--opt2", "Help 2")]
    child2 = MagicMock(spec=Target)
    child2.option_args.return_value = [("--opt2", "Help 2"), ("--opt3", "Help 3")]

    composite = CompositeTarget([child1, child2], "composite", "", TargetScope.LOCAL)
    options = composite.option_args()

    assert len(options) == 3
    assert ("--opt1", "Help 1") in options
    assert ("--opt2", "Help 2") in options
    assert ("--opt3", "Help 3") in options


@pytest.mark.parametrize(
    "child_allows_pass, expected_composite_allows_pass",
    [
        ([True, True], True),
        ([True, False], True),
        ([False, False], False),
    ],
)
def test_composite_target_allows_pass_args(
    child_allows_pass, expected_composite_allows_pass
):
    """Test CompositeTarget.allows_pass_args logic"""
    child1 = MagicMock(spec=Target)
    child1.allows_pass_args.return_value = child_allows_pass[0]
    child2 = MagicMock(spec=Target)
    child2.allows_pass_args.return_value = child_allows_pass[1]

    composite = CompositeTarget([child1, child2], "composite", "", TargetScope.LOCAL)
    assert composite.allows_pass_args() == expected_composite_allows_pass


def test_composite_target_pass_handler():
    """Test CompositeTarget.pass_handler joining"""
    child1 = MagicMock(spec=Target)
    child1.pass_handler.return_value = "handler1"
    child2 = MagicMock(spec=Target)
    child2.pass_handler.return_value = None
    child3 = MagicMock(spec=Target)
    child3.pass_handler.return_value = "handler3"

    composite = CompositeTarget(
        [child1, child2, child3], "composite", "", TargetScope.LOCAL
    )
    assert composite.pass_handler() == "handler1,handler3"



"""fprime.fpp.common:

Common implementations for FPP tool wrapping.

@author mstarch
"""

import itertools
import subprocess
from pathlib import Path
import shutil
import sys
import os
from typing import Dict, List, Tuple

from fprime.common.error import FprimeException
from fprime.fbuild.builder import Build
from fprime.fbuild.target import ExecutableAction, TargetScope


class FppMissingSupportFiles(FprimeException):
    """FPP is missing its support files and cannot run"""

    def __init__(self, file):
        super().__init__(
            f"Can not find {file}. Does current directory define any FPP files?"
        )


class FppUtility(ExecutableAction):
    """Action built around executing FPP

    When executing FPP, the utilities often require a set of input or dependency FPP files as well as a locations file
    that represents the location of the FPP constructs in the system. These inputs are then supplied to the FPP utility
    in a similar order across utilities. This action executes these utilities via a subprocess using the command line
    format:

    <utility> <user supplied args> [<fpp imports>] <locations file> <fpp inputs>

    Some fpp utilities distinguish between import files (--import flag) and source files. Those utilities should set
    the imports_as_sources flag to False. This will cause the utility to pass the import files as --import arguments.
    If imports_as_sources is True, the import files are passed as inputs just like source files.
    """

    def __init__(self, name, imports_as_sources=True):
        """Construct this utility with the supplied name

        Args:
            name: name of the fpp utility to run
        """
        super().__init__(TargetScope.LOCAL)
        self.utility = name
        self.imports_as_sources = imports_as_sources

    def is_supported(self, _=None, __=None):
        """Returns whether this utility is supported"""
        return bool(shutil.which(self.utility))

    @staticmethod
    def get_locations_file(builder: Build) -> Path:
        """Returns the location of the FPP locations file

        Returns a path to the FPP locations index within the build cache managed by the supplied build object.

        Args:
            builder: build object to use

        Return:
            path to the locations index
        """
        locations_path = Path(builder.build_dir) / "locs.fpp"
        if not locations_path.exists():
            raise FppMissingSupportFiles(locations_path)
        return locations_path

    @staticmethod
    def get_fpp_inputs(builder: Build, context: Path) -> Tuple[List[Path], List[Path]]:
        """Return the necessary inputs to an FPP run to forward to fpp utilities

        Returns two types of FPP files input into FPP utilities: the FPP files associated with the given module and the
        FPP files included by those files (e.g. commands.fpp). These files are output in a file in the build cache
        directory.

        Args:
            builder: build object used to look for output file
            context: context path of module containing the FPP files

        Return:
            tuple of two lists: module source FPP files and included FPP files
        """
        cache_location = builder.get_build_cache_path(context)
        import_file = cache_location / "fpp-import-list"
        source_file = cache_location / "fpp-source-list"
        if not import_file.exists():
            raise FppMissingSupportFiles(import_file)
        if not source_file.exists():
            raise FppMissingSupportFiles(source_file)
        with open(import_file, "r") as file_handle:
            import_list = [
                Path(token) for token in file_handle.read().split(";") if token != ""
            ]
        with open(source_file, "r") as file_handle:
            source_list = [
                Path(token) for token in file_handle.read().split(";") if token != ""
            ]
        return (import_list, source_list)

    def execute(
        self, builder: Build, context: Path, args: Tuple[Dict[str, str], List[str]]
    ):
        """Execute the fpp utility

        Executes the FPP utility wrapped by this class. Executes it with respect to the context path supplied and using
        the builder object. Supplies arguments in args to the utility.

        Args:
            builder: build object to run the utility with
            context: context path of module FPP is running on
            args: extra arguments to supply to the utility
        """
        # First refresh the cache but only if it detects it needs too

        if not self.is_supported():
            print(
                f"[ERROR] Cannot find executable: {self.utility}.",
                file=sys.stderr,
            )
            return 1

        builder.refresh()

        # Read files and arguments
        locations = self.get_locations_file(builder)
        imports, sources = self.get_fpp_inputs(builder, context)

        combined_env = os.environ.copy()
        combined_env.update(builder.settings.get("environment", {}))

        if not sources:
            print("[WARNING] No FPP sources found in this module.")

        # Build the input argument list
        input_args = []
        if self.imports_as_sources:
            input_args.extend(
                str(item) for item in itertools.chain([locations] + imports + sources)
            )
        else:
            input_args.extend(["-i", ",".join(map(str, imports))] if imports else [])
            input_args.extend(
                str(item) for item in itertools.chain([locations] + sources)
            )

        user_args = args[1]
        app_args = [self.utility] + user_args + input_args
        if builder.cmake.verbose:
            print(f"[FPP] '{' '.join(app_args)}'")
        return subprocess.run(
            app_args, cwd=context, capture_output=False, env=combined_env
        ).returncode



"""fprime.fbuild.target: build target support

Contains the supporting definitions for build targets. These targets are used to run various parts of the build and may
contain build system targets (e.g. CMake target invokers), and miscellaneous targets that perform other actions.

@author lestarch
"""

import functools
import itertools
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Dict, List, Set, Tuple

from .enumerator import BuildTargetEnumerator
from .types import BuildType, NoSuchTargetException

TargetContext = List[str]


class TargetScope(Enum):
    """Scoping for target execution: GLOBAL, LOCAL

    GLOBAL targets trigger top-level (global) build system targets. LOCAL targets trigger per-directory build system
    targets. BOTH represents a target that can operate in both LOCAL and GLOBAL mode one at a time. When registering
    a BOTH targets, the system will create a local target and a global target and register those. These targets differ
    in both scope and flags as the GLOBAL target receives the flag "--all" added to its list.
    """

    GLOBAL = 0x1
    LOCAL = 0x2
    BOTH = GLOBAL | LOCAL


class ExecutableAction(ABC):
    """Executable action not declaring a formal mnemonic, description, etc.

    Some steps in the execution of a composite target need to execute "actions lite" or anonymous targets. Things that
    have an execute method but are only executable through other targets. This class can be derived to create that
    without generating all the normal target metadata.
    """

    def __init__(self, scope: TargetScope):
        """Set scope of this action"""
        self.original_context = None
        self.scope = scope

    @abstractmethod
    def is_supported(self, builder: "Build", context_path: Path):
        """Is this target supported via the given contextual path"""
        pass

    @abstractmethod
    def execute(
        self,
        builder: "Build",
        context: Path,
        args: Tuple[Dict[str, str], List[str], Dict[str, bool]],
    ):
        """Executes the given targe with the given contextual path"""

    def option_args(self) -> List[Tuple[str, str]]:
        """List of option arguments handled by this target

        Option flags are switches that are not allowed arguments. The current design defaults the value to False unless
        the switch is supplied. This function is expected to return a list of pairs of switch flag and description help
        text. e.g.
        [
            (--turn-on, "Turns on a switch"),
        ]

        Returns:
            list of tuples containing the paired flag and description
        """
        return []

    def allows_pass_args(self):
        """Target allows pass-through arguments"""
        return False

    def pass_handler(self):
        """Handler of pass-through args"""
        return None

    def __repr__(self):
        """Representation"""
        return f"{self.__class__.__name__}"


class EnumeratedAction(ExecutableAction):
    """Action that derives targets for a given context path using an enumeration"""

    def __init__(
        self, scope: TargetScope, build_target_enumerator: BuildTargetEnumerator
    ):
        """Set scope of this action"""
        super().__init__(scope)
        self.build_target_enumerator = build_target_enumerator

    def is_supported(self, builder: "Build", context_path: Path):
        """Is supported by the list of build target names

        Checks if the build target names supplied will support this target. Is overridden by subclasses.

        Args:
            builder: builder to check if this action is supported
            context: contextual path to check

        Return:
            True if supported false otherwise
        """
        enumerated_targets: TargetContext = self.build_target_enumerator.enumerate(
            builder, context_path
        )
        return self.any_supported(builder, enumerated_targets)

    def execute(
        self,
        builder: "Build",
        context: Path,
        args: Tuple[Dict[str, str], List[str], Dict[str, bool]],
    ):
        """Executes the given target"""
        self.original_context = context
        enumerated_targets: TargetContext = self.build_target_enumerator.enumerate(
            builder, context
        )
        return self.execute_all(builder, enumerated_targets, args)

    def execute_all(
        self,
        builder: "Build",
        context: TargetContext,
        args: Tuple[Dict[str, str], List[str], Dict[str, bool]],
    ):
        """Executes all targets supplied via the context list

        This method is passed the enumerated context path, receiving context as a list of strings representing
        enumerated context as specified via self.build_target_enumerator.enumerate.

        Args:
            builder: builder to execute the target
            context: contextual path to execute the target
            args: arguments to pass to the target

        Returns:
            List of results from executing the target
        """
        pass

    def any_supported(self, builder: "Build", context: TargetContext) -> bool:
        """Is supported by the list of build target names

        Checks if the build target names supplied will support this target. Is overridden by subclasses.

        Args:
            builder: builder to check if this action is supported
            context: contextual path to check

        Return:
            True if supported false otherwise
        """
        return len(context) > 0


class Target(EnumeratedAction):
    """Generic build target base class

    A target can be specified by the user using a mnemonic and flags. The mnemonic is the command typed in by the user,
    and the flags allow the user to remember fewer mnemonics by changing the build target using a modifier. Each build
    target is available in certain build types.

    Targets can be global, using the GlobalTarget base class. Global targets don't use contextual information to modify
    the target, but apply to the whole deployment. Note: global targets are also engaged at the deployment level should
    that be the context.

    Targets may also be local. These targets use context information to figure out what to build. This allows for one
    target to represent a class of targets. i.e. build can be used as a local target to build any given sub directory.
    """

    ALL_TARGETS = []

    def __init__(
        self,
        mnemonic: str,
        desc: str,
        scope: TargetScope,
        build_type: BuildType = None,
        build_target_enumerator: BuildTargetEnumerator = None,
        flags: set = None,
    ):
        """Constructs a build target and registers it as one of the global targets

        As part of the construction of a Target it is registered as part of the targets available to be run by
        fprime-util. Targets defined as both global and local are wrapped in delegating targets (one for each scope) and
        those delegates are registered. This is for brevity in definition of these targets. The flag "--all" is added to
        global targets to distinguish them

        Args:
            mnemonic:    mnemonic used to engage build targets. Is not unique, but mnemonic + flags must be.
            desc:        help description of this build target
            build_types: supported build types for target. Defaults to [BuildType.BUILD_NORMAL, BuildType.BUILD_TESTING]
            build_target_enumerator: Enumerates build targets available in the given context path
            flags:       flags used to uniquely identify build targets who share logical mnemonics. Defaults to None.
            cmake:       cmake target override to handle oddly named cmake targets
        """
        super().__init__(scope=scope, build_target_enumerator=build_target_enumerator)
        self.mnemonic = mnemonic
        self.desc = desc
        self.build_type = (
            build_type if build_type is not None else BuildType.BUILD_NORMAL
        )
        self.flags = flags if flags is not None else set()

    @classmethod
    def register_target(cls, target: "Target"):
        """Registers the target"""
        cls.ALL_TARGETS.append(target)

    def __repr__(self):
        """Representation"""
        return f"{self.__class__.__name__}({str(self)})"

    def __str__(self):
        """Makes this target into a string"""
        return self.config_string(self.mnemonic, self.flags)

    @staticmethod
    def config_string(mnemonic, flags):
        """Converts a mnemonic and set of flags to string

        Args:
            mnemonic: mnemonic of the target
            flags: set of flags to pair with mnemonic
        Returns:
            string of format "mnemonic --flag1 --flag2 ..."
        """
        flag_string = " ".join([f"--{flag}" for flag in flags])
        flag_string = f" {flag_string}" if flag_string else ""
        return f"{mnemonic}{flag_string}"

    @classmethod
    def get_all_possible_flags(cls) -> Set[str]:
        """Gets list of all targets' flags used

        Returns:
            List of targets supported by the system
        """
        return functools.reduce(
            lambda agg, item: agg.union(item.flags), cls.get_all_targets(), set()
        )

    @classmethod
    def get_all_targets(cls) -> List["Target"]:
        """Gets list of all targets registered

        Returns:
            List of targets supported by the system
        """
        return cls.ALL_TARGETS

    @classmethod
    def get_target(cls, mnemonic: str, flags: Set[str]) -> "Target":
        """Gets the actual build target given the parsed namespace

        Using the global list of build targets and the flags supplied to the namespace, attempt to determine which build
        targets can be used. If more than one are found, then generate exception.

        Args:
            mnemonic: mnemonic of command to look for
            flags:    flags to narrow down target

        Returns:
            single matching target
        """
        matching = [
            target
            for target in cls.get_all_targets()
            if target.mnemonic == mnemonic and flags == target.flags
        ]
        if not matching:
            msg = f"Could not find target '{cls.config_string(mnemonic, flags)}'"
            raise NoSuchTargetException(msg)
        assert len(matching) == 1, "Conflicting targets specified in code"
        return matching[0]


class CompositeTarget(Target):
    """Target whose execution is a composition of other targets"""

    def __init__(self, composite_targets: List["Target"], *args, **kwargs):
        """Constructor setting child targets"""
        super().__init__(*args, **kwargs)
        self.targets = composite_targets

    def __repr__(self):
        """So we can see what it delegated to"""
        return f"{self.__class__.__name__}[{', '.join([target.__repr__() for target in self.targets])}]"

    def is_supported(self, builder: "Build", context: Path):
        """Any of the targets supported by the list of build target names

        Checks if the build target names supplied will support this target. Is overridden by subclasses.

        Args:
            builder: builder to check if this action is supported
            context: contextual path to check

        Return:
            True if supported false otherwise
        """
        # Supported only if all steps supported
        return functools.reduce(
            lambda sum, target: sum and target.is_supported(builder, context),
            self.targets,
            True,
        )

    def option_args(self):
        """Returns the set of option arguments"""
        return list(
            set(
                itertools.chain.from_iterable(
                    [target.option_args() for target in self.targets]
                )
            )
        )

    def allows_pass_args(self):
        """Pass args allowed if any child allows it"""
        return functools.reduce(
            lambda sum, target: sum or target.allows_pass_args(), self.targets, False
        )

    def pass_handler(self):
        """Pass handler as , separated list"""
        handlers = [
            target.pass_handler() for target in self.targets if target.pass_handler()
        ]
        return ",".join(handlers)

    def execute(self, *args, **kwargs):
        """Execute the composite target"""
        for child in self.targets:
            # Composite actions must override scope as a delegator may have acted to change the scope
            old_scope = child.scope
            try:
                child.scope = self.scope
                child.execute(*args, **kwargs)
            finally:
                child.scope = old_scope


class BuildSystemTarget(Target):
    """Target whose execution invokes a command within the build system"""

    def execute_all(
        self,
        builder: "Build",
        context: TargetContext,
        args: Tuple[Dict[str, str], List[str], Dict[str, bool]],
    ):
        for build_target in context:
            if builder.is_verbose():
                print(f"[INFO] Building: {build_target}")
            self.execute_one(builder, build_target, args)

    def execute_one(
        self,
        builder: "Build",
        build_target: str,
        args: Tuple[Dict[str, str], List[str], Dict[str, bool]],
    ):
        """Execute a build target

        Executes a target within the build system. This will execute the target by calling into the build system.
        Context is supplied such that the system can match local targets to the global target list.

        Args:
            builder: builder to execute target with
            build_target: build target to execute
            args: make system arguments directly supplied
        """
        # This assert guards against the old style calls through this function. It will catch when the refactor was
        # not completed correctly.
        assert not isinstance(build_target, Path), "Build target shall no be a path"

        # Execute the build target
        builder.execute_build_target(build_target, self.original_context, args[0])



"""fprime.fbuild.gcovr: coverage target implementations"""

import itertools
import shutil
import subprocess
import sys
import os
from pathlib import Path
from typing import Dict, List, Tuple, Union

from .enumerator import BuildTargetEnumerator
from .check import CheckTarget
from .target import CompositeTarget, ExecutableAction, Target, TargetScope


def _get_project_path(builder: "Build", module: Union[str, Path]) -> Path:
    """Get the project path from a given module string

    Get the project path for a given module or path. Paths are assumed to be contexts and returned verbatim. Modules are
    converted to a project string and then that is returned.

    Args:
        builder: builder object
        module: module string or project path
    """
    # If this module is path then return it as project path already
    if isinstance(module, Path):
        return module
    # Calculate the project path from the module name
    project_root = builder.get_settings(
        "project_root", builder.get_settings("framework_path", builder.build_dir.parent)
    )
    return Path(project_root) / module.replace("_", "/")


def _using_root(builder: "Build", context: Path, scope: TargetScope):
    """Are we using repository root for things"""
    return builder.is_project_root(context) or scope == TargetScope.GLOBAL


class Gcovr(ExecutableAction):
    """Target invoking `gcovr` utility to calculate coverage

    `gcovr` is a utility used to calculate coverage recursively. It is a wrapper around the gcov executable and patches
    several problems with the utility. This target runs `gcovr` and places the output in a newly created coverage
    directory. arguments are supplied into the `gcovr` execution, however; the default execution is always supplied the
    --txt coverage/summary.txt --txt stdout flags.
    """

    EXECUTABLE = "gcovr"

    def is_supported(self, builder: "Build", context: Path):
        """Is supported by the list of build target names

        Checks if the build target names supplied will support this target. Is overridden by subclasses.

        Args:
            builder: builder to check if this action is supported
            context: contextual path to check

        Return:
            True if supported false otherwise
        """
        return bool(shutil.which(self.EXECUTABLE))

    def execute(
        self,
        builder: "Build",
        context: Path,
        args: Tuple[Dict[str, str], List[str], Dict[str, bool]],
    ):
        """Executes the gcovr target"""
        if not shutil.which(self.EXECUTABLE):
            print(
                f"[ERROR] Cannot find executable: {self.EXECUTABLE}. Unable to run coverage report.",
                file=sys.stderr,
            )
            return
        build_cache_resolved = Path(builder.build_dir).resolve()

        make_args, pass_through_args, options = args

        # Handle jobs argument
        jobs_args = []
        jobs = make_args.get("--jobs", make_args.get("-j", None))
        if jobs is not None:
            jobs_args = ["-j", str(jobs)]

        include_all = options["--all-sources"]
        include_comp_ac = include_all or options["--comp-ac"]
        include_port_ac = include_all or options["--port-ac"]
        include_type_ac = include_all or options["--type-ac"]
        include_test_ac = include_all or options["--test-ac"]
        include_test = include_all or options["--test-sources"]
        build_cache_exclusion_filter_bases = [
            None if include_comp_ac else ".*ComponentAc.[ch]pp",
            None if include_port_ac else ".*PortAc.[ch]pp",
            None if include_type_ac else ".*SerializableAc.[ch]pp",
            None if include_type_ac else ".*ArrayAc.[ch]pp",
            None if include_type_ac else ".*EnumAc.[ch]pp",
            None if include_test_ac else ".*/.*GTestBase.[ch]pp",
            None if include_test_ac else ".*/.*TesterBase.[ch]pp",
            None if include_test_ac else ".*/.*TesterHelpers.[ch]pp",
            None if include_test_ac else ".*/.*StateMachineAc.[ch]pp",
        ]
        raw_source_exclusion_filter_bases = [
            None if include_test else ".*/test/.*",
            None if include_test else ".*/GTest/.*",
            None if include_test else "test/.*",
        ]

        build_cache_exclusion_filter_bases = filter(
            lambda item: item is not None, build_cache_exclusion_filter_bases
        )

        raw_source_exclusion_filter_bases = filter(
            lambda item: item is not None, raw_source_exclusion_filter_bases
        )

        exclusion_filter_bases = [
            ["--exclude", f"{build_cache_resolved}/{exclusion}"]
            for exclusion in build_cache_exclusion_filter_bases
        ] + [
            ["--exclude", f"{exclusion}"]
            for exclusion in raw_source_exclusion_filter_bases
        ]

        build_cache_path = (
            builder.build_dir
            if _using_root(builder, context, self.scope)
            else builder.get_build_cache_path(context)
        ).resolve()

        coverage_output_dir = context.resolve() / "coverage"
        coverage_output_dir.mkdir(exist_ok=True)
        project_root = builder.get_settings(
            "project_root",
            builder.get_settings("framework_path", builder.build_dir.parent.parent),
        ).resolve()
        filter_path = (
            Path(project_root).resolve()
            if _using_root(builder, context, self.scope)
            else _get_project_path(builder, context)
        ).resolve()
        framework_path = builder.get_settings(
            "framework_path", builder.build_dir.parent.parent
        )
        combined_env = os.environ.copy()
        combined_env.update(builder.settings.get("environment", {}))
        # gcovr is an unhappy beast
        cli_args = (
            [
                "gcovr",
                "-r",
                str(project_root),
                str(build_cache_path),  # For efficiency in searching on modules
            ]
            + jobs_args
            + list(itertools.chain.from_iterable(exclusion_filter_bases))
            + [
                "--exclude",
                f"{framework_path}/Autocoders",
                "--exclude",
                f"{framework_path}/gtest",
                "--exclude",
                f"{framework_path}/STest",
                "--filter",
                f"{filter_path}",
                "--filter",
                f"{build_cache_path}",
                "--print-summary",
                "--txt",
                f"{coverage_output_dir}/summary.txt",
                "--html-details",
                f"{coverage_output_dir}/coverage{'-all' if self.scope == TargetScope.GLOBAL else ''}.html",
            ]
        )
        cli_args.extend(pass_through_args)

        if builder.cmake.verbose:
            joined_args = "' '".join(cli_args)
            print(f"[INFO] Running \"'{ joined_args }'\"")
        # gcovr must run in the ac_temporary_path or html details cannot find the Ac files
        subprocess.call(cli_args, env=combined_env)

    def option_args(self):
        """Option arguments"""
        return [
            ("--all-sources", "[coverage only] Include all sources in coverage"),
            ("--comp-ac", "[coverage only] Include component autocode in coverage"),
            ("--port-ac", "[coverage only] Include port autocode in coverage"),
            ("--type-ac", "[coverage only] Include data type autocode in coverage"),
            ("--test-ac", "[coverage only] Include data test autocode in coverage"),
            ("--test-sources", "[coverage only] Include unit test sources in coverage"),
        ]

    def allows_pass_args(self):
        """Gcovr allows pass-through arguments"""
        return True

    def pass_handler(self):
        """Pass handler"""
        return self.EXECUTABLE


class GcovrTarget(CompositeTarget):
    """Target specific to gcovr

    The gcovr target is a composite target that builds upon an existing check target to perform gcovr work. In addition,
    it must support extra arguments as we pass these to the gcovr executable.
    """

    def __init__(
        self,
        scope: TargetScope,
        build_target_enumerators: List[BuildTargetEnumerator],
        *args,
        **kwargs,
    ):
        """Constructor setting child targets"""
        check_target = CheckTarget(
            scope=scope,
            build_target_enumerators=build_target_enumerators,
            *args,
            **kwargs,
        )
        # Reuse the enumerator for tests
        gcovr_target = Gcovr(scope)
        super().__init__(
            [check_target, gcovr_target],
            scope=scope,
            *args,
            **kwargs,
        )



"""
This file defines the structures for the annotatable elements in the fpp AST. These
are separate from the converters itself such that it may be possible to write
converters from other languages to these Python data structures.

If extending this utility, the first step would be to add a new class for the data
structure for any new element. Then, a converter is required to turn an input AST into
these structures.
"""


class Module:
    """
    Attributes:
        ast: The AST of the module
        name: The name of the module
        members: The members of the module
        preannot: The preannotations of the module
        postannot: The postannotations of the module
        qf: The qualified name of the module
    """

    def __init__(self, ast):
        self.ast = ast
        self.name = None
        self.members = []

        self.preannot = None
        self.postannot = None
        self.qf = None


class Constant:
    """
    Attributes:
        ast: The AST of the constant
        id: The id (LHS) of the constant
        value: The value of the constant
        preannot: The preannotations of the constant
        postannot: The postannotations of the constant
        qf: The qualified name of the constant
    """

    def __init__(self, ast):
        self.ast = ast
        self.id = None
        self.value = None

        self.preannot = None
        self.postannot = None
        self.qf = None


class InstanceSpec:
    """
    Attributes:
        ast: The AST of the instance spec
        name: The name of the instance spec
        visibility: The visibility of the instance spec (private, normal)
        preannot: The preannotations of the instance spec
        postannot: The postannotations of the instance spec
        qf: The qualified name of the instance spec
    """

    def __init__(self, ast):
        self.ast = ast
        self.name = None
        self.visibility = None

        self.preannot = None
        self.postannot = None
        self.qf = None


class ComponentInst:
    """
    Attributes:
        ast: The AST of the component instance
        name: The name of the component instance
        component_name: The name of the component
        base_id: The base id of the component instance
        queue_size: The queue size of the component instance
        stack_size: The stack size of the component instance
        cpu: The cpu of the component instance
        priority: The priority of the component instance
        phases: The phases of the component instance
        preannot: The preannotations of the component instance
        postannot: The postannotations of the component instance
        qf: The qualified name of the component instance
    """

    def __init__(self, ast):
        self.ast = ast
        self.name = None
        self.component_name = None
        self.base_id = None
        self.queue_size = None
        self.stack_size = None
        self.cpu = None
        self.priority = None
        self.phases = {
            "configObjects": None,
            "configComponents": None,
            "readParameters": None,
            "configConstants": None,
            "tearDownComponents": None,
            "startTasks": None,
            "stopTasks": None,
            "freeThreads": None,
        }

        self.preannot = None
        self.postannot = None
        self.qf = None


class Port:
    """
    Attributes:
        ast: The AST of the port
        name: The name of the port
        type: The type of the port
        preannot: The preannotations of the port
        postannot: The postannotations of the port
        qf: The qualified name of the port
    """

    def __init__(self, ast):
        self.ast = ast
        self.name = None
        self.parameters = []

        self.preannot = None
        self.postannot = None
        self.qf = None


class TopologyImport:
    """
    Attributes:
        ast: The AST of the topology import
        name: The name of the topology import
        preannot: The preannotations of the topology import
        postannot: The postannotations of the topology import
        qf: The qualified name of the topology import
    """

    def __init__(self, ast):
        self.ast = ast
        self.name = None

        self.preannot = None
        self.postannot = None
        self.qf = None


class ConnectionGraph:
    """
    Attributes:
        ast: The AST of the connection graph
        name: The name of the connection graph
        connections: The connections of the connection graph
        type: The type of the connection graph
        preannot: The preannotations of the connection graph
        postannot: The postannotations of the connection graph
        qf: The qualified name of the connection graph
    """

    def __init__(self, ast):
        self.ast = ast
        self.name = None
        self.connections = []
        self.type = None

        self.preannot = None
        self.postannot = None
        self.qf = None


class Topology:
    """
    Attributes:
        ast: The AST of the topology
        name: The name of the topology
        members: The members of the topology
        preannot: The preannotations of the topology
        postannot: The postannotations of the topology
        qf: The qualified name of the topology
    """

    def __init__(self, ast):
        self.ast = ast
        self.name = None
        self.members = []

        self.preannot = None
        self.postannot = None
        self.qf = None



import os
from pathlib import Path
import shutil
import json
import fprime.fpp.utils.fpp_to_json.fpp_interface as fpp


def qualifier_calculator(qualifier_JSON):
    """
    Calculate the qualified name for a variable, instance, etc

    Args:
        qualifier_JSON: The JSON AST for the qualifier
    """
    next_idx = qualifier_JSON["e"]["AstNode"]["data"]

    path = ""

    if "ExprDot" in next_idx:
        path += qualifier_calculator(next_idx["ExprDot"])
    elif "ExprIdent" in next_idx:
        path += next_idx["ExprIdent"]["value"]

    return path + "." + qualifier_JSON["id"]["AstNode"]["data"]


def value_parser(value_JSON):
    """
    Parse a value from the JSON AST. This could be parsing integers, strings, arrays, etc.

    Args:
        value_JSON: The JSON AST for the value
    """
    checker = value_JSON["AstNode"]["data"]

    if "Unqualified" in checker:
        return checker["Unqualified"]["name"]
    elif "Qualified" in checker:
        qf = (
            checker["Qualified"]["qualifier"]["AstNode"]["data"]["Unqualified"]["name"]
            + "."
            + checker["Qualified"]["name"]["AstNode"]["data"]
        )
        return qf
    elif "ExprIdent" in checker:
        return checker["ExprIdent"]["value"]
    elif "ExprDot" in checker:
        return qualifier_calculator(checker["ExprDot"])
    else:
        return parse_constant(value_JSON)


def Binops(binop):
    """
    Convert binop string name to actual operator

    Args:
        binop: The binop string name
    """
    if binop == "Add":
        return "+"
    elif binop == "Sub":
        return "-"
    elif binop == "Mul":
        return "*"
    elif binop == "Div":
        return "/"

    raise Exception("Invalid binop")


def parse_binop(constant_JSON):
    """
    Parse a binary operation from the JSON AST

    Args:
        constant_JSON: The JSON AST for the binary operation
    """
    idx_LHS = 1
    idx_RHS = 2
    binop = ""

    constant_JSON = constant_JSON["ExprBinop"]
    binop_LHS = constant_JSON[f"e{idx_LHS}"]["AstNode"]["data"]

    if "ExprBinop" in binop_LHS:
        binop += parse_binop(binop_LHS)
    else:
        binop += value_parser(constant_JSON[f"e{idx_LHS}"])

    return (
        binop
        + " "
        + Binops(list(constant_JSON["op"].keys())[0])
        + " "
        + value_parser(constant_JSON[f"e{idx_RHS}"])
    )


def parse_constant(constant_JSON):
    """
    Parse a constant from the JSON AST

    Args:
        constant_JSON: The JSON AST for the constant
    """
    constant_Value_JSON = constant_JSON["AstNode"]["data"]

    if "ExprLiteralString" in constant_Value_JSON:
        return '"' + constant_Value_JSON["ExprLiteralString"]["value"] + '"'
    elif "ExprLiteralInt" in constant_Value_JSON:
        return constant_Value_JSON["ExprLiteralInt"]["value"]
    elif "ExprLiteralFloat" in constant_Value_JSON:
        return constant_Value_JSON["ExprLiteralFloat"]["value"]
    elif "ExprLiteralBool" in constant_Value_JSON:
        return list(constant_Value_JSON["ExprLiteralBool"]["value"].keys())[0]
    elif "ExprIdent" in constant_Value_JSON:
        return constant_Value_JSON["ExprIdent"]["value"]
    elif "ExprArray" in constant_Value_JSON:
        return parse_array(constant_Value_JSON)
    elif "ExprStruct" in constant_Value_JSON:
        return parse_struct(constant_Value_JSON)
    elif "ExprDot" in constant_Value_JSON:
        return qualifier_calculator(constant_Value_JSON["ExprDot"])
    elif "ExprBinop" in constant_Value_JSON:
        return parse_binop(constant_Value_JSON)
    else:
        raise Exception("Invalid constant type")


def parse_array(constant_JSON):
    """
    Parse an array from the JSON AST

    Args:
        constant_JSON: The JSON AST for the array
    """
    arrayOpen = "["
    constant_Value_JSON = constant_JSON["ExprArray"]["elts"]

    for i in range(len(constant_Value_JSON)):
        if "ExprArray" not in constant_Value_JSON[i]["AstNode"]["data"]:
            arrayOpen = arrayOpen + parse_constant(constant_Value_JSON[i])
            if i is not len(constant_Value_JSON) - 1:
                arrayOpen = arrayOpen + ", "
        else:
            arrayOpen = (
                arrayOpen
                + parse_array(constant_Value_JSON[i]["AstNode"]["data"])
                + ", "
            )

    return arrayOpen + "]"


def parse_struct(constant_JSON):
    """
    Parse a struct from the JSON AST

    Args:
        constant_JSON: The JSON AST for the struct
    """
    structOpen = "{\n"

    for param in constant_JSON["ExprStruct"]["members"]:
        structOpen = (
            structOpen
            + "    "
            + param["AstNode"]["data"]["name"]
            + " = "
            + value_parser(param["AstNode"]["data"]["value"])
            + ",\n"
        )

    return structOpen + "}"


def openFppFile(path):
    if not os.path.isabs(path):
        path = str(Path(path).resolve())

    pathDir = os.path.dirname(path)
    fileBasename = os.path.basename(path)
    folderName = f"/{fileBasename.split('.')[0]}Cache"

    pathToFolder = pathDir + folderName

    if not os.path.exists(pathToFolder):
        try:
            os.mkdir(pathToFolder)
        except OSError as e:
            raise Exception("Creation of the directory %s failed" % (pathToFolder))

    os.chdir(pathToFolder)

    if not os.path.exists("fpp-ast.json"):
        fpp.fpp_to_json(path)

    # parse json
    with open("fpp-ast.json", "r") as f:
        AST = json.load(f)

    os.chdir(pathDir)

    shutil.rmtree(pathToFolder, ignore_errors=True)

    return AST



from enum import Enum
from typing import List

from fprime.common.error import FprimeException


class InvalidBuildCacheException(FprimeException):
    """An exception indicating a build cache"""

    def __init__(self, message, build_cache=""):
        super().__init__(message)
        self.cache = build_cache


class UnableToDetectProjectException(FprimeException):
    """An exception indicating a build cache"""


class NoSuchTargetException(FprimeException):
    """Could not find a matching build target"""


class NoSuchToolchainException(FprimeException):
    """Could not find a matching build target"""


class AmbiguousToolchainException(FprimeException):
    """Could not find a matching build target"""


class InvalidBuildTypeException(FprimeException):
    """An exception indicating a build type do not exit"""


class MissingBuildCachePath(FprimeException):
    """An exception indicating that a path in the build cache is missing"""


class BuildType(Enum):
    """
    An enumeration used to represent the various build types used to build fprime. These types can support different
    types of targets underneath. i.e. the unit-test build may build unit test executables.
    """

    """ Normal build normal binaries for a deployment mapping to CMake 'Release'"""  # pylint: disable=W0105
    BUILD_NORMAL = (0,)
    """ Testing build allowing unit testing mapping to CMake 'Testing'"""  # pylint: disable=W0105
    BUILD_TESTING = 1
    """ FPP locations build """
    BUILD_FPP_LOCS = 2
    """ User custom build """
    BUILD_CUSTOM = 3

    def get_suffix(self):
        """Get the suffix of a directory supporting this build"""
        if self == BuildType.BUILD_NORMAL:
            return ""
        if self == BuildType.BUILD_TESTING:
            return "-ut"
        msg = f"{self.name} is not a supported build type"
        raise InvalidBuildTypeException(msg)

    def get_cmake_build_type(self):
        """Get the suffix of a directory supporting this build"""
        if self == BuildType.BUILD_NORMAL:
            return "Release"
        if self == BuildType.BUILD_TESTING:
            return "Testing"
        if self == BuildType.BUILD_FPP_LOCS:
            return "Release"
        if self == BuildType.BUILD_CUSTOM:
            return "Custom"
        msg = f"{self.name} is not a supported build type"
        raise InvalidBuildTypeException(msg)

    @staticmethod
    def get_public_types() -> List["BuildType"]:
        """Returns public build types"""
        return [BuildType.BUILD_NORMAL, BuildType.BUILD_TESTING]



from pathlib import Path
from unittest.mock import MagicMock

import pytest
from fprime.fbuild.enumerator import (
    BasicBuildTargetEnumerator,
    CliBuildTargetEnumerator,
    DesignateTargetAction,
    MultiBuildTargetEnumerator,
    RecursiveMultiBuildTargetEnumerator,
    SpecificBuildTargetEnumerator,
)
from fprime.fbuild.types import MissingBuildCachePath

# Path to the test data directory
ENUMERATOR_DATA_PATH = Path(__file__).parent / "enumerator_data"


def test_basic_enumeration():
    """Test basic enumeration with and without a suffix"""
    # Setup mock builder
    builder_mock = MagicMock()
    builder_mock.cmake.get_cmake_module.return_value = "TestModule"

    # Test without suffix
    enumerator = BasicBuildTargetEnumerator()
    context_path = ENUMERATOR_DATA_PATH / "BasicBuildTargetEnumerator"
    assert enumerator.enumerate(builder_mock, context_path) == ["TestModule"]
    builder_mock.cmake.get_cmake_module.assert_called_once_with(
        context_path, builder_mock.build_dir
    )


def test_basic_enumeration_with_suffix():
    """Test basic enumeration with a suffix"""
    # Setup mock builder
    builder_mock = MagicMock()
    builder_mock.cmake.get_cmake_module.return_value = "TestModule"
    # Test with suffix
    enumerator_suffix = BasicBuildTargetEnumerator(target_suffix="test")
    context_path = ENUMERATOR_DATA_PATH / "BasicBuildTargetEnumerator"
    assert enumerator_suffix.enumerate(builder_mock, context_path) == [
        "TestModule_test"
    ]
    builder_mock.cmake.get_cmake_module.assert_called_once_with(
        context_path, builder_mock.build_dir
    )


def test_multi_build_target_enumerator_default():
    """Test MultiBuildTargetEnumerator with default file"""
    data_path = ENUMERATOR_DATA_PATH / "MultiBuildTargetEnumerator"
    builder_mock = MagicMock()
    builder_mock.get_build_cache_path.return_value = data_path

    # Test with default build-targets.fprime-util
    enumerator = MultiBuildTargetEnumerator()
    expected_targets = ["targetA", "targetB", "targetC"]
    assert enumerator.enumerate(builder_mock, data_path) == expected_targets


def test_multi_build_target_enumerator_tests_file():
    """Test MultiBuildTargetEnumerator with tests.fprime-util"""
    data_path = ENUMERATOR_DATA_PATH / "MultiBuildTargetEnumerator"
    builder_mock = MagicMock()
    builder_mock.get_build_cache_path.return_value = data_path

    # Test with tests.fprime-util
    test_enumerator = MultiBuildTargetEnumerator(
        build_file_name=MultiBuildTargetEnumerator.TEST_BUILD_TARGETS_FILE
    )
    expected_test_targets = ["testTargetX", "testTargetY"]
    assert test_enumerator.enumerate(builder_mock, data_path) == expected_test_targets


def test_multi_build_target_enumerator_fallback():
    """Test MultiBuildTargetEnumerator fallback"""
    # Test fallback by pointing to a directory without the file
    fallback_path = ENUMERATOR_DATA_PATH / "BasicBuildTargetEnumerator"
    builder_mock = MagicMock()
    builder_mock.get_build_cache_path.return_value = fallback_path
    fail_safe_mock = MagicMock()
    fail_safe_mock.enumerate.return_value = ["fallback_target"]
    fallback_enumerator = MultiBuildTargetEnumerator(
        fail_safe_enumerator=fail_safe_mock
    )
    assert fallback_enumerator.enumerate(builder_mock, fallback_path) == [
        "fallback_target"
    ]
    fail_safe_mock.enumerate.assert_called_once_with(builder_mock, fallback_path)


def test_multi_build_target_enumerator_fallback_exception():
    """Test MultiBuildTargetEnumerator fallback on exception"""
    context_path = Path("/path/to/context")
    builder_mock = MagicMock()
    builder_mock.get_build_cache_path.side_effect = MissingBuildCachePath(context_path)

    fail_safe_mock = MagicMock()
    fail_safe_mock.enumerate.return_value = ["fallback_target"]
    fallback_enumerator = MultiBuildTargetEnumerator(
        fail_safe_enumerator=fail_safe_mock
    )

    assert fallback_enumerator.enumerate(builder_mock, context_path) == [
        "fallback_target"
    ]
    fail_safe_mock.enumerate.assert_called_once_with(builder_mock, context_path)


def test_recursive_multi_build_target_enumerator():
    """Tests the recursive enumeration of targets"""
    root_path = ENUMERATOR_DATA_PATH / "RecursiveMultiBuildTargetEnumerator"
    builder_mock = MagicMock()

    # For our test setup, the 'build cache' is the directory itself.
    builder_mock.get_build_cache_path.side_effect = lambda path: path

    enumerator = RecursiveMultiBuildTargetEnumerator()
    found_targets = enumerator.enumerate(builder_mock, root_path)

    expected_targets = {
        "root_target1",
        "root_target2",
        "sub1_target1",
        "sub2_target1",
        "sub2_1_target1",
    }
    # Using set for order-independent comparison
    assert set(found_targets) == expected_targets


def test_recursive_multi_build_target_enumerator_custom_file():
    """Tests the recursive enumeration with a custom test file"""
    root_path = ENUMERATOR_DATA_PATH / "RecursiveMultiBuildTargetEnumerator"
    builder_mock = MagicMock()

    # For our test setup, the 'build cache' is the directory itself.
    builder_mock.get_build_cache_path.side_effect = lambda path: path

    # Create a MultiBuildTargetEnumerator that looks for the test file
    test_file_enumerator = MultiBuildTargetEnumerator(
        build_file_name=MultiBuildTargetEnumerator.TEST_BUILD_TARGETS_FILE
    )
    # Create the recursive enumerator with the custom enumerator
    enumerator = RecursiveMultiBuildTargetEnumerator(
        directory_enumerator=test_file_enumerator
    )

    found_targets = enumerator.enumerate(builder_mock, root_path)

    expected_targets = {
        "root_test1",
        "sub2_test1",
        "sub2_test2",
    }
    assert set(found_targets) == expected_targets


def test_recursive_enumerator_fallback_on_file_not_found():
    """Test recursive enumerator falls back when helper returns empty"""
    # Point to a directory that exists but does not contain target/sub-directory files
    path = ENUMERATOR_DATA_PATH / "BasicBuildTargetEnumerator"
    builder_mock = MagicMock()
    builder_mock.get_build_cache_path.return_value = path
    # When the recursive search yields no targets, it falls back to the basic
    # enumerator, which will call get_cmake_module. We mock its return value.
    builder_mock.cmake.get_cmake_module.return_value = "fallback_target"

    enumerator = RecursiveMultiBuildTargetEnumerator()
    # Since enumerate_helper returns [], the fallback enumerator should be called
    assert enumerator.enumerate(builder_mock, path) == ["fallback_target"]
    # Verify that the fallback was indeed triggered
    builder_mock.cmake.get_cmake_module.assert_called_once_with(
        path, builder_mock.build_dir
    )


def test_specific_build_target_enumerator_constructor():
    """Test SpecificBuildTargetEnumerator returns constructor-provided targets"""
    targets = ["specific_target1", "specific_target2"]
    enumerator = SpecificBuildTargetEnumerator(build_targets=targets)
    # The builder and context path are not used by this enumerator
    assert enumerator.enumerate(None, None) == targets


def test_specific_build_target_enumerator_setter():
    """Test SpecificBuildTargetEnumerator's setter overrides constructor targets"""
    initial_targets = ["initial1"]
    override_targets = ["override1", "override2"]

    enumerator = SpecificBuildTargetEnumerator(build_targets=initial_targets)
    assert enumerator.enumerate(None, None) == initial_targets

    enumerator.set_build_targets(override_targets)
    assert enumerator.enumerate(None, None) == override_targets


def test_cli_build_target_enumerator():
    """Test CliBuildTargetEnumerator integrates with DesignateTargetAction"""
    # Reset the designees for a clean test
    DesignateTargetAction._DESIGNEES = []

    # 1. Test Registration and enumeration
    cli_enumerator = CliBuildTargetEnumerator()
    assert cli_enumerator in DesignateTargetAction._DESIGNEES

    # 2. Test Target Setting via Action
    action = DesignateTargetAction(option_strings=["--target"], dest="target")
    parser_mock = MagicMock()
    namespace = MagicMock()
    values = ["cli_target1", "cli_target2"]

    # Simulate argparse calling the action, which should call the real `set_build_targets`
    action(parser_mock, namespace, values)

    # Verify that the namespace is updated as expected
    assert namespace.target is True
    # Verify that a subsequent call to enumerate returns the values set by the action
    assert cli_enumerator.enumerate(None, None) == values



"""
(test) fprime.util.cookiecutter_wrapper - new_deployment():
Test adding a new deployment to a project using cookiecutter.

This test verifies the behavior of creating deployments in non-root directories,
specifically testing the automatic detection and adjustment of include paths.
"""

import os
import shutil
import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch, MagicMock

from fprime.fbuild.builder import Build
from fprime.util.cookiecutter_wrapper import new_deployment


class TestNewDeployment(unittest.TestCase):
    """Test the new_deployment function"""

    def setUp(self):
        """Set up a temporary directory structure for testing"""
        # Create a temporary directory to simulate a project
        self.temp_dir = tempfile.mkdtemp()
        self.project_root = Path(self.temp_dir) / "project_root"
        self.project_root.mkdir()

        # Create a subdirectory to simulate a Deployments directory
        self.deployments_dir = self.project_root / "Deployments"
        self.deployments_dir.mkdir()

        # Create a CMakeLists.txt file in the project root
        with open(self.project_root / "CMakeLists.txt", "w") as f:
            f.write("# Project CMakeLists.txt\n")

        # Create a settings.ini file in the project root
        with open(self.project_root / "settings.ini", "w") as f:
            f.write("[fprime]\n")
            f.write("project_root = .\n")
            f.write(f"framework_path = {self.project_root}\n")

        # Save the original working directory
        self.original_cwd = os.getcwd()

    def tearDown(self):
        """Clean up temporary directory"""
        # Change back to the original working directory
        os.chdir(self.original_cwd)

        # Remove the temporary directory
        shutil.rmtree(self.temp_dir)

    @patch("fprime.util.cookiecutter_wrapper.cookiecutter")
    @patch("fprime.util.cookiecutter_wrapper.register_with_cmake")
    def test_new_deployment_in_root_dir(self, mock_register, mock_cookiecutter):
        """Test creating a deployment in the project root directory"""
        # Change to the project root directory
        os.chdir(self.project_root)

        # Mock the build object
        mock_build = MagicMock(spec=Build)
        mock_build.get_settings.return_value = self.project_root
        mock_build.cmake_root = self.project_root

        # Mock the parsed arguments
        mock_args = MagicMock()
        mock_args.force = False
        mock_args.overwrite = False

        # Mock the cookiecutter function to return a path
        mock_cookiecutter.return_value = str(self.project_root / "MyDeployment")

        # Call the function
        result = new_deployment(mock_build, mock_args)

        # Verify the result
        self.assertEqual(result, 0)

        # Verify that cookiecutter was called with the correct arguments
        mock_cookiecutter.assert_called_once()
        args, kwargs = mock_cookiecutter.call_args
        self.assertIn("extra_context", kwargs)
        self.assertEqual(kwargs["extra_context"], {})  # No include path prefix

    @patch("fprime.util.cookiecutter_wrapper.cookiecutter")
    @patch("fprime.util.cookiecutter_wrapper.register_with_cmake")
    def test_new_deployment_in_subdir(self, mock_register, mock_cookiecutter):
        """Test creating a deployment in a subdirectory"""
        # Change to the Deployments directory
        os.chdir(self.deployments_dir)

        # Mock the build object
        mock_build = MagicMock(spec=Build)
        mock_build.get_settings.return_value = self.project_root
        mock_build.cmake_root = self.project_root

        # Mock the parsed arguments
        mock_args = MagicMock()
        mock_args.force = False
        mock_args.overwrite = False

        # Mock the cookiecutter function to return a path
        mock_cookiecutter.return_value = str(self.deployments_dir / "MyDeployment")

        # Call the function
        result = new_deployment(mock_build, mock_args)

        # Verify the result
        self.assertEqual(result, 0)

        # Verify that cookiecutter was called with the correct arguments
        mock_cookiecutter.assert_called_once()
        args, kwargs = mock_cookiecutter.call_args
        self.assertIn("extra_context", kwargs)
        self.assertIn("__include_path_prefix", kwargs["extra_context"])
        self.assertEqual(
            kwargs["extra_context"]["__include_path_prefix"], "Deployments/"
        )


if __name__ == "__main__":
    unittest.main()



"""fprime.fbuild.enumerator: build module enumeration strategies

This file contains the necessary enumeration strategies for build system targets. Enumeration is used to determine
what targets are available in the given contextual path (working directory). The results of the enumeration are then
passed to the target for execution.

Note: this file was created with the help of generative AI.

"""

from abc import ABC, abstractmethod
from argparse import Action
from pathlib import Path
from typing import List

from .types import MissingBuildCachePath


class DesignateTargetAction(Action):
    """This class, when used as the argparse action will set the target of all DesignatedBuildSystemTarget

    argparse.Action actions can be used to handle custom behavior when parsing commands. This class will use the
    --target flag to specify the build target on all known DesignatedBuildSystemTarget that have registered with this
    class.
    """

    _DESIGNEES = []

    @classmethod
    def register_designee(cls, designee):
        """Register designee to this action"""
        cls._DESIGNEES.append(designee)

    def __call__(self, parser, namespace, values, option_string=None):
        """Required __call__ function triggered by the parse"""

        # For each designee, update the build system target
        for designee in self._DESIGNEES:
            designee.set_build_targets(values)
        # The build target detection looks for true/false flags to be set. Mimic this by setting 'target' to True
        setattr(namespace, "target", True)


class BuildTargetEnumerator(ABC):
    """Abstract base class for target enumeration strategies."""

    @abstractmethod
    def enumerate(self, builder: "Build", context_path: Path) -> List[str]:
        """Enumerates build targets available in the given context path."""
        raise NotImplementedError()


class BasicBuildTargetEnumerator(BuildTargetEnumerator):
    """A basic enumerator that just returns the context given."""

    def __init__(self, target_suffix: str = ""):
        self.target_suffix = target_suffix

    def enumerate_helper(self, builder: "Build", context_path: Path) -> List[str]:
        """Enumeration helper without error handling"""
        module = builder.cmake.get_cmake_module(context_path, builder.build_dir)
        if module and self.target_suffix:
            return [module + "_" + self.target_suffix]
        elif module:
            return [module]
        return [self.target_suffix]

    def enumerate(self, builder: "Build", context_path: Path) -> List[str]:
        """Enumerate through conversion to cmake module name directly"""
        return self.enumerate_helper(builder, context_path)


class MultiBuildTargetEnumerator(BuildTargetEnumerator):
    """Enumerates targets from the 'build-targets.fprime-util' file."""

    BUILD_TARGETS_FILE = "build-targets.fprime-util"
    TEST_BUILD_TARGETS_FILE = "tests.fprime-util"

    def __init__(
        self,
        build_file_name: str = BUILD_TARGETS_FILE,
        fail_safe_enumerator: BuildTargetEnumerator = None,
    ):
        """Construct the enumerator with the name of the file"""
        self.build_file_name = build_file_name
        self.fail_safe_enumerator = (
            BasicBuildTargetEnumerator()
            if fail_safe_enumerator is None
            else fail_safe_enumerator
        )

    def enumerate_helper(self, builder: "Build", context_path: Path) -> List[str]:
        """Enumeration helper without error handling"""
        build_cache_path = builder.get_build_cache_path(context_path)
        build_targets_file = build_cache_path / self.build_file_name
        with open(build_targets_file, "r") as file_handle:
            enumerated_targets = [line.strip() for line in file_handle.readlines()]
        return enumerated_targets

    def enumerate(self, builder: "Build", context_path: Path) -> List[str]:
        """Enumerates build targets, falling back to the context if not found."""
        try:
            return self.enumerate_helper(builder, context_path)
        # should an error occur, delegate to a basic enumeration strategy
        except (MissingBuildCachePath, FileNotFoundError):
            return self.fail_safe_enumerator.enumerate(builder, context_path)


class RecursiveMultiBuildTargetEnumerator(BuildTargetEnumerator):
    """Recursively enumerates targets."""

    SUBDIRECTORIES_FILE = "sub-directories.fprime-util"

    def __init__(self, directory_enumerator: "MultiBuildTargetEnumerator" = None):
        """Construct the enumerator with the name of the file"""
        self.directory_enumerator = (
            directory_enumerator
            if directory_enumerator is not None
            else MultiBuildTargetEnumerator()
        )

    def enumerate_helper(self, builder: "Build", context_path: Path) -> List[str]:
        """Enumeration helper without error handling"""
        enumerated_targets = []
        try:
            # Get local targets
            try:
                enumerated_targets.extend(
                    self.directory_enumerator.enumerate_helper(builder, context_path)
                )
            except FileNotFoundError:
                pass  # No local targets, continue to recursion

            # Recurse through subdirectories
            build_cache_path = builder.get_build_cache_path(context_path)
            sub_directories_file = build_cache_path / self.SUBDIRECTORIES_FILE
            with open(sub_directories_file, "r") as file_handle:
                for sub_dir in file_handle.readlines():
                    # Each line is a new path relative to the current context
                    # We need to construct the full path for the recursive call
                    full_sub_dir_path = (build_cache_path / sub_dir.strip()).resolve()
                    enumerated_targets.extend(
                        self.enumerate_helper(builder, full_sub_dir_path)
                    )
        # Any error should return the result of the last update to enumerate_targets hance passing
        except (MissingBuildCachePath, FileNotFoundError):
            pass
        return enumerated_targets

    def enumerate(self, builder: "Build", context_path: Path) -> List[str]:
        """Recursively enumerates build targets."""
        enumerated_targets = self.enumerate_helper(builder, context_path)
        enumerated_targets = (
            enumerated_targets
            if enumerated_targets
            else self.directory_enumerator.enumerate(builder, context_path)
        )
        return enumerated_targets


class SpecificBuildTargetEnumerator(BuildTargetEnumerator):
    """Enumerates a specific singular build target"""

    def __init__(self, build_targets: List[str] = None):
        """Construct the enumerator with the name of the build target"""
        self.build_targets = build_targets if build_targets is not None else []

    def set_build_targets(self, build_targets: List[str]):
        """Set the build targets"""
        self.build_targets = build_targets

    def enumerate(self, builder: "Build", context_path: Path) -> List[str]:
        """Enumerates exactly a build target"""
        return self.build_targets


class CliBuildTargetEnumerator(SpecificBuildTargetEnumerator):
    """Allows users to specify a build target on the command line"""

    def __init__(self):
        super().__init__(build_targets=None)
        DesignateTargetAction.register_designee(self)



"""fprime.fbuild.code_formatter

Wrapper for clang-format utility.

@author thomas-bc
"""

import re
import os
import shutil
import subprocess
from pathlib import Path
from typing import Dict, List, Tuple

from fprime.fbuild.target import ExecutableAction, TargetScope

# clang-format will try to format everything it is given - restrict for the time being
ALLOWED_EXTENSIONS = [
    ".cpp",
    ".c++",
    ".cxx",
    ".cc",
    ".c",
    ".hpp",
    ".h++",
    ".hxx",
    ".hh",
    ".h",
]


class ClangFormatter(ExecutableAction):
    """Class encapsulating the clang-format logic for fprime-util"""

    def __init__(self, executable: str, style_file: "Path", options: Dict):
        super().__init__(TargetScope.LOCAL)
        self.executable = executable
        self.style_file = style_file
        self.backup = options.get("backup", False)
        self.verbose = options.get("verbose", False)
        self.quiet = options.get("quiet", False)
        self.check = options.get("check", False)
        self.validate_extensions = options.get("validate_extensions", True)
        self.allowed_extensions = ALLOWED_EXTENSIONS.copy()
        self._files_to_format: List[Path] = []

    def is_supported(self, _=None, __=None) -> bool:
        return bool(shutil.which(self.executable))

    def allow_extension(self, file_ext: str) -> None:
        """Add a file extension str to the list of allowed extension"""
        self.allowed_extensions.append(file_ext)

    def stage_file(self, filepath: Path) -> None:
        """Request ClangFormatter to consider the file for formatting.
        If the file exists and its extension matches a known C/C++ format,
        it will be passed to clang-format when the execute() function is called.

        Args:
            filepath (str): file path to file to be staged.
        """
        if not filepath.is_file():
            if self.verbose:
                print(f"[INFO] Skipping {filepath} : is not a file.")
        elif self.validate_extensions and (
            filepath.suffix not in self.allowed_extensions
        ):
            if self.verbose:
                print(
                    f"[INFO] Skipping {filepath} : unrecognized C/C++ file extension "
                    f"('{filepath.suffix}'). Use --allow-extension or --force."
                )
        else:
            self._files_to_format.append(filepath)

    def execute(
        self, builder: "Build", context: "Path", args: Tuple[Dict[str, str], List[str]]
    ):
        """Execute clang-format on the files that were staged.

        Args:
            builder (Build): build object to run the utility with
            context (Path): context path of module clang-format can run on if --module is provided
            args (Tuple[Dict[str, str], List[str]]): extra arguments to supply to the utility
        """
        combined_env = os.environ.copy()
        combined_env.update(builder.settings.get("environment", {}))

        if len(self._files_to_format) == 0:
            print("[INFO] No files were formatted.")
            return 0
        if not self.style_file.is_file():
            print(
                f"[ERROR] No .clang-format file found in {self.style_file.parent}. "
                "Override location with --pass-through --style=file:<path>."
            )
            return 1
        # Backup files unless --no-backup is requested or running only a --check
        if self.backup and not self.check:
            for file in self._files_to_format:
                shutil.copy2(file, file.parent / f"{file.stem}.bak{file.suffix}")
        pass_through = args[1]
        clang_args = [
            self.executable,
            "-i",
            f"--style=file",
            *(["--verbose"] if not self.quiet else []),
            *(["--dry-run", "--Werror"] if self.check else []),
            *pass_through,
            *self._files_to_format,
        ]
        if self.verbose:
            print("[INFO] Clang format executable:")
            print(f"[INFO]    {self.executable}")
            print("[INFO] Clang format arguments:")
            print(f"[INFO]    {clang_args[1:]}")
            print("[INFO] Clang format style file:")
            print(f"[INFO]    {self.style_file}")
        status = subprocess.run(clang_args, env=combined_env)
        return status.returncode



"""Cookie cutter wrapper used to template out components"""

import glob
import os
import sys

from typing import TYPE_CHECKING
from contextlib import contextmanager
from pathlib import Path

from cookiecutter.exceptions import OutputDirExistsException
from cookiecutter.main import cookiecutter

from fprime.common.utils import confirm, check_path_is_within_fprime_module
from fprime.fbuild.builder import Build
from fprime.fbuild.cmake import CMakeExecutionException
from fprime.fpp.impl import fpp_generate_implementation
from fprime.util.file_util import get_directory_path_relative_to_root

if TYPE_CHECKING:
    import argparse


def run_impl(build: Build, source_path: Path):
    """Run implementation of files in source_path"""
    if not confirm("Generate implementation files?"):
        return False
    print("Refreshing cache and generating implementation files...")

    with suppress_stdout():
        fpp_generate_implementation(build, source_path, source_path, True, False)

    # Path(source_path).
    file_list = glob.glob(f"{source_path}/*.template.*pp", recursive=False)
    for filename in file_list:
        new_filename = filename.replace(".template", "")
        os.rename(filename, new_filename)

    return True


@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


def find_nearest_cmake_file(component_dir: Path, cmake_root: Path, proj_root: Path):
    """Find the nearest CMake file, i.e. CMakeLists.txt or project.cmake

    The "nearest" file is defined as the closest parent that is not the project root CMakeLists.txt.
    If none is found, the same procedure is run from the deployment directory and includes the project
    root this time. If nothing is found, None is returned.

    In short the following in order of preference:
     - Any Component Parent
     - Any Deployment Parent
     - project.cmake
     - None

    Args:
        component_dir: directory of new component
        deployment: deployment directory
        proj_root: project root directory

    Returns:
        path to CMakeLists.txt or None
    """
    test_path = component_dir.parent
    # First iterate from where we are, then from the deployment to find the nearest CMakeList.txt nearby
    for test_path, end_path in [(test_path, proj_root), (cmake_root, proj_root.parent)]:
        while proj_root is not None and test_path != proj_root.parent:
            project_file = test_path / "project.cmake"
            if project_file.is_file():
                return project_file
            cmake_list_file = test_path / "CMakeLists.txt"
            if cmake_list_file.is_file():
                return cmake_list_file
            test_path = test_path.parent
    return None


def new_component(build: Build, parsed_args: "argparse.Namespace"):
    """Uses cookiecutter for making new components"""

    if (
        check_path_is_within_fprime_module(path=Path.cwd(), is_component=True)
        and not parsed_args.force
    ):
        print(
            "[ERROR] Wrong location. Cannot create component within an existing component."
            " Use --force to override."
        )
        return 1

    try:
        proj_root = build.get_settings("project_root", None)

        # Checks if component_cookiecutter is set in settings.ini file, else uses local component_cookiecutter template as default
        if (
            build.get_settings("component_cookiecutter", None) is not None
            and build.get_settings("component_cookiecutter", None) != "default"
        ):
            source = build.get_settings("component_cookiecutter", None)
            print(f"[INFO] Cookiecutter source: {source}")
        else:
            source = (
                os.path.dirname(__file__)
                + "/../cookiecutter_templates/cookiecutter-fprime-component"
            )
            print("[INFO] Cookiecutter source: using builtin")

        # Use current working directory name as default namespace, unless at project root
        extra_context = {}
        if not proj_root.samefile(Path.cwd()):
            extra_context["component_namespace"] = Path.cwd().name

        gen_path = Path(cookiecutter(source, extra_context=extra_context)).resolve()

        if proj_root is None:
            print(
                f"[INFO] Created component directory without adding to build system nor generating implementation {gen_path}"
            )
            return 0
        # Attempt to register to CMakeLists.txt or project.cmake
        register_with_cmake(
            gen_path,
            Path(proj_root).resolve(),
            build.cmake_root,
        )
        # Attempt implementation
        if not run_impl(build, gen_path):
            print(
                f"[INFO] Did not generate implementations for {gen_path}. Please do so manually."
            )
            return 0

        print("[INFO] Created new component and generated initial implementations.")
        return 0
    except OutputDirExistsException as out_directory_error:
        print(f"{out_directory_error}", file=sys.stderr)
    except CMakeExecutionException as exc:
        print(f"[ERROR] Failed to create component. {exc}", file=sys.stderr)
    except FileNotFoundError as e:
        print(
            f"{e}. Permission denied to write to the directory.",
            file=sys.stderr,
        )
        return 1
    except OSError as ose:
        print(f"[ERROR] {ose}")
    return 1


def new_deployment(build: Build, parsed_args: "argparse.Namespace"):
    """Creates a new deployment using cookiecutter"""

    if check_path_is_within_fprime_module(Path.cwd()) and not parsed_args.force:
        print(
            "[ERROR] Wrong location. Cannot create deployment within an existing component or deployment"
        )
        return 1

    # Checks if deployment_cookiecutter is set in settings.ini file, else uses local install template as default
    if (
        build.get_settings("deployment_cookiecutter", None) is not None
        and build.get_settings("deployment_cookiecutter", None) != "default"
    ):
        source = build.get_settings("deployment_cookiecutter", None)
        print(f"[INFO] Cookiecutter source: {source}")
    else:
        source = (
            os.path.dirname(__file__)
            + "/../cookiecutter_templates/cookiecutter-fprime-deployment"
        )
        print("[INFO] Cookiecutter: using builtin template for new deployment")

    # Determine the include path prefix based on the current directory
    extra_context = {}
    rel_path = get_directory_path_relative_to_root(build)

    if rel_path:
        extra_context["__include_path_prefix"] = f"{rel_path}/"
        print(
            f"[INFO] Creating a deployment in a subdirectory. Include paths will be prefixed with '{rel_path}/'"
        )

    try:
        gen_path = Path(
            cookiecutter(
                source,
                extra_context=extra_context,
                overwrite_if_exists=parsed_args.overwrite,
            )
        ).resolve()
        # Attempt to register to CMakeLists.txt or project.cmake
        register_with_cmake(
            gen_path,
            Path(build.get_settings("project_root", None)).resolve(),
            build.cmake_root,
        )

    except OutputDirExistsException as out_directory_error:
        print(
            f"{out_directory_error}. Use --overwrite to overwrite (will not delete non-generated files).",
            file=sys.stderr,
        )
        return 1
    except FileNotFoundError as e:
        print(
            f"{e}. Permission denied to write to the directory.",
            file=sys.stderr,
        )
        return 1
    print(f"[INFO] New deployment successfully created: {gen_path}")
    return 0


def new_subtopology(build: Build, parsed_args: "argparse.Namespace"):
    """Creates a new subtopology using cookiecutter"""
    # Checks if subtopology_cookiecutter is set in settings.ini file, else uses local install template as default
    if (
        build.get_settings("subtopology_cookiecutter", None) is not None
        and build.get_settings("subtopology_cookiecutter", None) != "default"
    ):
        source = build.get_settings("subtopology_cookiecutter", None)
        print(f"[INFO] Cookiecutter source: {source}")
    else:
        source = (
            os.path.dirname(__file__)
            + "/../cookiecutter_templates/cookiecutter-fprime-subtopology"
        )
        print("[INFO] Cookiecutter: using builtin template for new subtopology")
    try:
        gen_path = Path(
            cookiecutter(source, overwrite_if_exists=parsed_args.overwrite)
        ).resolve()
        # Attempt to register to CMakeLists.txt or project.cmake
        register_with_cmake(
            gen_path,
            Path(build.get_settings("project_root", None)).resolve(),
            build.cmake_root,
        )

    except OutputDirExistsException as out_directory_error:
        print(
            f"{out_directory_error}. Use --overwrite to overwrite (will not delete non-generated files).",
            file=sys.stderr,
        )
        return 1
    except FileNotFoundError as e:
        print(
            f"{e}. Permission denied to write to the directory.",
            file=sys.stderr,
        )
        return 1
    print(f"[INFO] New subtopology successfully created: {gen_path}")
    return 0


def new_module(build: Build, parsed_args: "argparse.Namespace"):
    """Creates a new F' project"""

    source = (
        os.path.dirname(__file__)
        + "/../cookiecutter_templates/cookiecutter-fprime-module"
    )
    try:
        gen_path = Path(
            cookiecutter(
                source,
                overwrite_if_exists=parsed_args.overwrite,
                output_dir=parsed_args.path,
            )
        ).resolve()
        # Attempt to register to CMakeLists.txt or project.cmake
        register_with_cmake(
            gen_path,
            Path(build.get_settings("project_root", None)).resolve(),
            build.cmake_root,
        )
    except OutputDirExistsException as out_directory_error:
        print(
            f"{out_directory_error}. Use --overwrite to overwrite (will not delete non-generated files).",
            file=sys.stderr,
        )
        return 1
    except FileNotFoundError as e:
        print(
            f"{e}. Permission denied to write to the directory.",
            file=sys.stderr,
        )
        return 1
    return 0


def register_with_cmake(gen_path: Path, proj_root: Path, cmake_root: Path):
    cmake_file = find_nearest_cmake_file(gen_path, cmake_root, proj_root)
    if cmake_file is None or not add_to_cmake(
        cmake_file,
        gen_path.relative_to(cmake_file.parent),
        proj_root,
    ):
        print(
            f"[INFO] Could not register {gen_path} with build system. Please add it manually."
        )


def add_to_cmake(list_file: Path, comp_path: Path, project_root: Path = None):
    """Adds comp_path directory to CMakeLists.txt. If project_root is supplied,
    the logged path will be relative to the project root instead of absolute"""
    short_display_path = (
        list_file
        if project_root is None
        else project_root.name / list_file.relative_to(project_root)
    )
    print(f"[INFO] Found CMake file at '{short_display_path}'")
    with open(list_file, "r") as f:
        lines = f.readlines()

    addition = (
        'add_fprime_subdirectory("${CMAKE_CURRENT_LIST_DIR}/' + str(comp_path) + '/")\n'
    )
    if addition in lines:
        print("Already added to CMakeLists.txt")
        return True

    if not confirm(f"Add {comp_path} to {short_display_path} at end of file?"):
        return False

    # Handle case where the last line does not end with a newline
    if len(lines) > 0 and (not lines[-1].endswith("\n")):
        lines[-1] += "\n"

    lines.append(addition)
    with open(list_file, "w") as f:
        f.write("".join(lines))
    return True


def is_valid_name(word: str):
    invalid_characters = [
        "#",
        "%",
        "&",
        "{",
        "}",
        "/",
        "\\",
        "<",
        ">",
        "*",
        "?",
        " ",
        "$",
        "!",
        "'",
        '"',
        ":",
        "@",
        "+",
        "`",
        "|",
        "=",
        "-",
    ]
    for char in invalid_characters:
        if isinstance(word, str) and char in word:
            return char
        if not isinstance(word, str):
            raise ValueError("Incorrect usage of is_valid_name")
    return "valid"



"""fprime.fbuild.check: check target implementation

The 'check' target is designed to call CTest executable(s) to run tests. It is a composite target used to build and run
the test targets.
"""

import shutil
import subprocess
import sys
from typing import Tuple, Dict, List

from fprime.fbuild.target import (
    TargetContext,
    TargetScope,
    EnumeratedAction,
    CompositeTarget,
)

from fprime.fbuild.target import BuildSystemTarget
from .enumerator import BuildTargetEnumerator


class Check(EnumeratedAction):
    """Target invoking CTest executable to run tests"""

    EXECUTABLE = "ctest"

    def __init__(
        self, scope: TargetScope, build_target_enumerator: BuildTargetEnumerator = None
    ):
        """Initialize this action with the test enumerator"""
        super().__init__(scope=scope, build_target_enumerator=build_target_enumerator)

    def any_supported(self, builder: "Build", context: TargetContext):
        """Check if this target is supported in the given context

        This will determine if two conditions are met:
        1. CTest is available
        2. There are test targets to run
        """
        if not bool(shutil.which(self.EXECUTABLE)):
            print("[ERROR] CTest not found", file=sys.stderr)
            return False
        return len(context) > 0

    def execute_all(
        self,
        builder: "Build",
        context: TargetContext,
        args: Tuple[Dict[str, str], List[str], Dict[str, bool]],
    ):
        """Execute this target"""
        cli_args = [
            self.EXECUTABLE,
            "--test-dir",
            str(builder.build_dir),
            "--no-tests=error",
        ]
        make_args = args[0]

        # Jobs flag
        if "-j" in make_args or "--jobs" in make_args:
            cli_args.extend(
                ["--parallel", str(make_args.get("--jobs", make_args.get("-j", 1)))]
            )
        # Check for conditions that result in verbose output
        # 1. Explicitly verbose
        # 2. Context is not "all" (or .*)
        if builder.is_verbose() or (
            context and context != ["all"] and ".*" not in context
        ):
            cli_args.append("-V")
        # When not "all" append a regex to filter tests. .* works as a regex
        if context and context != ["all"]:
            test_regex = f"^({'|'.join(context)})$"
            cli_args.extend(["-R", test_regex])
        # Extend with "pass through" arguments
        cli_args.extend(args[1])
        # When verbose print the commands run
        if builder.is_verbose():
            joined = "' '".join(cli_args)
            print(f"[INFO] Running CTest: '{joined}'")

        # Check ensures errors result in this process erroring
        subprocess.run(cli_args, check=True)


class CheckTarget(CompositeTarget):
    """Target designed to build and run the tests"""

    def __init__(
        self,
        scope: TargetScope,
        build_target_enumerators: List[BuildTargetEnumerator],
        *args,
        **kwargs,
    ):
        """Constructor setting child targets"""
        build_target = BuildSystemTarget(
            scope=scope,
            build_target_enumerator=build_target_enumerators[0],
            *args,
            **kwargs,
        )
        check_action = Check(
            scope=scope, build_target_enumerator=build_target_enumerators[1]
        )
        super().__init__(
            composite_targets=[build_target, check_action],
            scope=scope,
            *args,
            **kwargs,
        )



from fprime.util.cookiecutter_wrapper import is_valid_name

name = "{{ cookiecutter.deployment_name }}"

if is_valid_name(name) != "valid":
    raise ValueError(
        f"Unacceptable deployment name: {name}. Do not use spaces or special characters"
    )



"""FPP tools to requirements file version check"""

import argparse
import sys
from pathlib import Path
import pkg_resources


class VersionException(Exception):
    pass


FPRIME_PIP_PACKAGES = [
    "fprime-tools",
    "fprime-gds",
    "fprime-fpp",
]


def get_version(package: str, requirements: Path):
    """Get the version as specified in the requirements file

    This will read all requirements from the requirements file and attempt to print the version of the package that is
    listed within. This can handle multiple styles of requirements. Firm requirements designated by a "==" and developer
    requirements designated with an "@".

    Args:
        package: name of package to look for
        requirements: path to requirements file to parse
    """
    with open(requirements, "r") as file_handle:
        matching_lines = [
            line.strip() for line in file_handle.readlines() if package in line
        ]
    if not matching_lines:
        msg = f"Could not find {package} in requirements file"
        raise VersionException(msg)
    valid_lines = [line for line in matching_lines if "==" in line or "@" in line]
    if not valid_lines:
        msg = f"{package} has inexact version, use '==' or '@' format. Found: {matching_lines}"
        raise VersionException(msg)

    # Collapse versions that match
    versions = list({line.split("==")[-1].split("@")[-1] for line in valid_lines})
    if len(versions) != 1:
        msg = f"Conflicting versions specified for {package}: {versions}"
        raise VersionException(msg)
    return versions[0]


def main():
    """Parses arguments and acts as entry point for the tool"""
    parser = argparse.ArgumentParser(
        description="Detects the required package using a requirements file"
    )
    parser.add_argument("package", help="Package to check for version")
    parser.add_argument(
        "requirements", help="Python formatted requirements file to parse"
    )

    try:
        args_ns = parser.parse_args()
        python_version = get_version(args_ns.package, Path(args_ns.requirements))

        # Add expected v at the front, if missing
        if "." in python_version and not python_version.lower().startswith("v"):
            python_version = f"v{python_version}"
        # Add in a g as a version-control tag for hash versions:
        elif "." not in python_version:
            python_version = f"g{python_version}"
        print(python_version)
    except (IOError, VersionException) as exc:
        print(f"[ERROR] Could not detect expected version: {exc}", file=sys.stderr)
        sys.exit(-1)
    sys.exit(0)


if __name__ == "__main__":
    main()
